#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np

from TransportMaps.Distributions.DistributionBase import ConditionalDistribution
from TransportMaps.Distributions.FrozenDistributions import GaussianDistribution

__all__ = ['ConditionallyGaussianDistribution',
           'MeanConditionallyGaussianDistribution']

nax = np.newaxis

class ConditionallyGaussianDistribution(ConditionalDistribution):
    r""" Multivariate Gaussian distribution :math:`\pi({\bf x}\vert{\bf y}) \sim \mathcal{N}(\mu({\bf y}), \Sigma({\bf y}))`

    Args:
      mu (:class:`Map<TransportMaps.Maps.Map>`): mean vector map
      sigma (:class:`Map<TransportMaps.Maps.Map>`): covariance matrix map
      precision (:class:`Map<TransportMaps.Maps.Map>`): precision matrix map
      coeffs (:class:`ndarray<numpy.ndarray>`): fix the coefficients :math:`{\bf y}`
    """
    def __init__(self, mu, sigma=None, precision=None, coeffs=None):
        if (sigma is not None) and (precision is not None):
            raise ValueError("The fields sigma and precision are mutually " +
                             "exclusive")
        if sigma is not None and mu.dim_in != sigma.dim_in:
            raise ValueError("The number of parameters must be the same for both " + \
                             "the map mu and the map sigma")
        if precision is not None and mu.dim_in != precision.dim_in:
            raise ValueError("The number of parameters must be the same for both " + \
                             "the map mu and the map precision")
        self._muMap = mu
        self._sigmaMap = sigma
        self._precisionMap = precision
        if sigma is not None:
            self._isSigmaOn = True
        else:
            self._isSigmaOn = False
        self._mu = None
        self._sigma = None
        self._precision = None
        self._pi = GaussianDistribution(np.zeros(mu.dim_out), np.eye(mu.dim_out))
        super(ConditionallyGaussianDistribution,self).__init__(mu.dim_out, mu.dim_in)
        self._coeffs = None
        self.coeffs = coeffs

    @property
    def pi(self):
        return self._pi
        
    @property
    def mu(self):
        return self._pi.mu
        
    @property
    def sigma(self):
        return self._pi.sigma

    @property
    def precision(self):
        return self._pi.inv_sigma
        
    @property
    def muMap(self):
        return self._muMap

    @property
    def sigmaMap(self):
        return self._sigmaMap

    @property
    def precisionMap(self):
        return self._precisionMap

    @property
    def coeffs(self):
        return self._coeffs

    @coeffs.setter
    def coeffs(self, coeffs):
        if coeffs is None:
            self._coeffs = None
        elif self._coeffs is None or np.any(self._coeffs != coeffs):
            # Set up Gaussian distribution
            mu = self._muMap.evaluate(coeffs[nax,:])[0,:]
            sigma = None
            precision = None
            if self._isSigmaOn:
                sigma = self._sigmaMap.evaluate(
                    coeffs[nax,:])[0,:,:] 
                self._precision = None
            else:
                precision = self._precisionMap.evaluate(
                    coeffs[nax,:])[0,:,:]
                self._sigma = None
            # Set up gradients
            try:
                self._grad_a_mu = self._muMap.grad_x(coeffs[nax,:])[0,:,:]
                if self._isSigmaOn:
                    self._grad_a_sigma = \
                        self._sigmaMap.grad_x(coeffs[nax,:])[0,:,:,:]
                    self._grad_a_precision = None
                else:
                    self._grad_a_precision = \
                        self._precisionMap.grad_x(coeffs[nax,:])[0,:,:,:]
                    self._grad_a_sigma = None
            except NotImplementedError:
                self._grad_a_c = None
                self._grad_a_T = None
            try:
                self._hess_a_mu = self._muMap.hess_x(
                    coeffs[nax,:])[0,:,:,:]
                if self._isSigmaOn:
                    self._hess_a_sigma = \
                        self._sigmaMap.hess_x(coeffs[nax,:])[0,:,:,:,:]
                    self._hess_a_precision = None
                else:
                    self._hess_a_precision = \
                        self._precisionMap.hess_x(coeffs[nax,:])[0,:,:,:,:]
                    self._hess_a_sigma = None
            except NotImplementedError:
                self._hess_a_c = None
                self._hess_a_T = None
            if self._pi is None:
                self._pi = GaussianDistribution(mu, sigma=sigma, precision=precision)
            else:
                self._pi.mu = mu
                if self._isSigmaOn:
                    self._pi.sigma = sigma
                else:
                    self._pi.precision = precision
            self._coeffs = coeffs

    @property
    def grad_a_mu(self):
        return self._grad_a_mu

    @property
    def grad_a_sigma(self):
        return self._grad_a_sigma

    @property
    def grad_a_precision(self):
        return self._grad_a_precision

    def rvs(self, m, y=None, **kwargs):
        r""" Generate :math:`m` samples from the distribution.

        Args:
          m (int): number of samples to generate
          y (:class:`ndarray<numpy.ndarray>` [:math:`d_y`]): conditioning values
            :math:`{\bf Y}={\bf y}`

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- :math:`m`
             :math:`d`-dimensional samples

        Raises:
          NotImplementedError: the method needs to be defined in the sub-classes
        """
        if self._coeffs is None:
            self.pi.mu = self._muMap.evaluate(y[nax,:])[0,:]
            if self._isSigmaOn:
                self.pi.sigma = self._sigmaMap.evaluate(y[nax,:])[0,:,:] 
            else:
                self.pi.precision = self._precisionMap.evaluate(y[nax,:])[0,:,:]
        return self.pi.rvs(m)

    def log_pdf(self, x, y, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          y (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y`] or :class:`ndarray<numpy.ndarray>` [:math:`d_y`]):
            conditioning values :math:`{\bf Y}={\bf y}`. In the second case one
            conditioning value is used for all the :math:`m` points :math:`{\bf x}`
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of :math:`\log\pi`
            at the ``x`` points.
        """
        if self._coeffs is None:
            m = x.shape[0]
            mu = self._muMap.evaluate(y)
            if self._isSigmaOn:
                sigma = self._sigmaMap.evaluate(y)
            else:
                precision = self._precisionMap.Evaluate(y)
            out = np.zeros(m)
            for i in range(m):
                self.pi.mu = mu[i]
                if self._isSigmaOn:
                    self.pi.sigma = sigma[i]
                else:
                    self.pi.precision = precision[i]
                out[i] = self.pi.log_pdf(
                    x[[i],:], params=params, idxs_slice=idxs_slice)
        else:
            out = self.pi.log_pdf(x, params=params, idxs_slice=idxs_slice)
        return out

    def grad_x_log_pdf(self, x, y, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\nabla_{\bf x,y} \log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          y (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y`]): conditioning values
            :math:`{\bf Y}={\bf y}`
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- values of
            :math:`\nabla_x\log\pi` at the ``x`` points.
        """
        if self._coeffs is None:
            m = x.shape[0]
            mu_all = self._muMap.evaluate(y)
            ga_mu_all = self._muMap.grad_x(y)
            if self._isSigmaOn:
                sigma_all = self._sigmaMap.evaluate(y)
                ga_sigma_all = self._sigmaMap.evaluate(y)
            else:
                precision_all = self._precisionMap.Evaluate(y)
                raise NotImplementedError("There are some parts missing for this gradient")
            out = np.zeros((m, x.shape[1]+y.shape[1]))
            for i in range(m):
                mu = mu_all[i]
                ga_mu = ga_mu_all[i]
                sigma = sigma_all[i]
                ga_sigma = ga_sigma_all[i]
                # Cholesky of sigma
                chol = scila.cho_factor(sigma)
                inv_sigma = scila.cho_solve(chol, np.eye(mu.shape))
                ga_inv_sigma = - np.einsum('ij,jkl,km->iml', inv_sigma, ga_sigma, inv_sigma)
                xmu = x[[i],:]-mu
                isgas = np.einsum('ij,jkl->ikl', inv_sigma, ga_inv_sigma)
                out[i,:x.shape[1]] = self.pi.grad_x_log_pdf(
                    x[[i],:], params=params, idxs_slice=idxs_slice)
                out[i,x.shape[1]:] = \
                    -.5 * ( np.einsum('ij,ik,...k->...j', -ga_mu, inv_sigma, xmu) + \
                            np.einsum('...i,ikl,...k->...l', xmu, ga_inv_sigma, xmu) + \
                            np.einsum('...i,ij,jk->...k', xmu, inv_sigma, -ga_mu) ) \
                    - .5 * np.einsum('iik->k', isgas) # Log determinant part
        else:
            raise NotImplementedError("To be done.")
        return out

class MeanConditionallyGaussianDistribution(ConditionalDistribution):
    r""" Multivariate Gaussian distribution :math:`\pi({\bf x}\vert{\bf y}) \sim \mathcal{N}(\mu({\bf y}), \Sigma)`

    Args:
      mu (:class:`Map<TransportMaps.Maps.Map>`): mean vector map
      sigma (:class:`ndarray<numpy.ndarray>`): covariance matrix map
      precision (:class:`ndarray<numpy.ndarray>`): precision matrix map
      coeffs (:class:`ndarray<numpy.ndarray>`): fix the coefficients :math:`{\bf y}`
    """
    def __init__(self, mu, sigma=None, precision=None, coeffs=None):
        if (sigma is not None) and (precision is not None):
            raise ValueError("The fields sigma and precision are mutually " +
                             "exclusive")
        self._muMap = mu
        self._mu = None
        self._pi = GaussianDistribution(np.zeros(mu.dim_out), sigma)
        super(MeanConditionallyGaussianDistribution,self).__init__(mu.dim_out, mu.dim_in)
        self._coeffs = None
        self.coeffs = coeffs

    @property
    def pi(self):
        return self._pi
        
    @property
    def mu(self):
        return self._pi.mu
        
    @property
    def sigma(self):
        return self._pi.sigma

    @property
    def precision(self):
        return self._pi.inv_sigma
        
    @property
    def muMap(self):
        return self._muMap

    @property
    def coeffs(self):
        return self._coeffs

    @coeffs.setter
    def coeffs(self, coeffs):
        if coeffs is None:
            self._coeffs = None
        elif self._coeffs is None or np.any(self._coeffs != coeffs):
            # Set up Gaussian distribution
            mu = self._muMap.evaluate(coeffs[nax,:])[0,:]
            self._pi.mu = mu
            self._coeffs = coeffs

    @property
    def grad_a_mu(self):
        return self._grad_a_mu

    def rvs(self, m, y=None, **kwargs):
        r""" Generate :math:`m` samples from the distribution.

        Args:
          m (int): number of samples to generate
          y (:class:`ndarray<numpy.ndarray>` [:math:`d_y`]): conditioning values
            :math:`{\bf Y}={\bf y}`

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- :math:`m`
             :math:`d`-dimensional samples

        Raises:
          NotImplementedError: the method needs to be defined in the sub-classes
        """
        if self._coeffs is None:
            self.pi.mu = self._muMap.evaluate(y[nax,:])[0,:]
        return self.pi.rvs(m)

    def log_pdf(self, x, y, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          y (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y`] or :class:`ndarray<numpy.ndarray>` [:math:`d_y`]):
            conditioning values :math:`{\bf Y}={\bf y}`. In the second case one
            conditioning value is used for all the :math:`m` points :math:`{\bf x}`
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of :math:`\log\pi`
            at the ``x`` points.
        """
        if self._coeffs is None:
            m = x.shape[0]
            mu = self._muMap.evaluate(y)
            out = np.zeros(m)
            for i in range(m):
                self.pi.mu = mu[i]
                out[i] = self.pi.log_pdf(
                    x[[i],:], params=params, idxs_slice=idxs_slice)
        else:
            out = self.pi.log_pdf(x, params=params, idxs_slice=idxs_slice)
        return out

    def grad_x_log_pdf(self, x, y, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\nabla_{\bf x,y} \log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          y (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y`]): conditioning values
            :math:`{\bf Y}={\bf y}`
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- values of
            :math:`\nabla_x\log\pi` at the ``x`` points.
        """
        raise NotImplementedError("Not implemented yet")


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

from TransportMaps.ObjectBase import TMO
from TransportMaps.Misc import deprecate
from TransportMaps.Distributions.Decomposable import \
    SequentialHiddenMarkovChainDistribution
from TransportMaps.Maps.TriangularTransportMapBase import \
    TriangularTransportMap, TriangularListStackedTransportMap
from TransportMaps.Maps.TransportMapBase import ListCompositeTransportMap

__all__ = ['Filter', 'Smoother']

class Filter(TMO):
    r""" Perform the on-line filtering of a sequential Hidded Markov chain.

    Given the prior distribution on the hyper-parameters :math:`\pi(\Theta)`,
    provides the functions neccessary to assimilate new pieces of data or
    missing data 
    (defined in terms of transition densities
    :math:`\pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k, \Theta \right)`
    and log-likelihoods
    :math:`\log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}, \Theta\right)`),
    to return the maps pushing forward :math:`\mathcal{N}(0,{\bf I})`
    to the filtering/forecast distributions
    :math:`\{\pi\left(\Theta, {\bf Z}_k \middle\vert {\bf y}_{0:k} \right)\}_k`.

    For more details see also :cite:`Spantini2017` and the
    `tutorial <example-sequential-stocvol-6d.html>`_.

    Args:
      pi_hyper (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        prior distribution on the hyper-parameters :math:`\pi(\Theta)`

    .. note:: This is a super-class. Part of its methods need to be implemented
      by sub-classes.
    """
    def __init__(self, pi_hyper=None):
        super(Filter,self).__init__()
        self._pi = SequentialHiddenMarkovChainDistribution([], [], pi_hyper)
        self._nsteps = -1
        self.F_list = [] # List of filtering maps

    @property
    def pi(self):
        return self._pi

    @property
    def nsteps(self):
        return self._nsteps

    @property
    def filtering_map_list(self):
        r""" Returns the maps :math:`\{ \widetilde{\mathfrak{M}}_k({\bf x}_\theta, {\bf x}_{k+1}) \}_{i=0}^{k-1}` pushing forward :math:`\mathcal{N}(0,{\bf I})` to the filtering/forecast distributions :math:`\{\pi\left(\Theta, {\bf Z}_k \middle\vert {\bf y}_{0:k} \right)\}_k`.

        The maps :math:`\widetilde{\mathfrak{M}}_k({\bf x}_\theta, {\bf x}_{k+1})`
        are defined as follows:

        .. math::

           \widetilde{\mathfrak{M}}_k({\bf x}_\theta, {\bf x}_{k+1}) = 
           \left[\begin{array}{l}
           \mathfrak{M}_0^\Theta \circ \cdots \circ \mathfrak{M}_{k}^\Theta ({\bf x}_\theta) \\
           \mathfrak{M}_k^1\left({\bf x}_\theta, {\bf x}_{k+1}\right)
           \end{array}\right] =
           \left[\begin{array}{l}
           \mathfrak{H}_{k}({\bf x}_\theta) \\
           \mathfrak{M}_k^1\left({\bf x}_\theta, {\bf x}_{k+1}\right)
           \end{array}\right]

        Returns:
          (:class:`list` of :class:`TransportMap<TransportMaps.Maps.TransportMap>`) -- list of transport maps :math:`\widetilde{\mathfrak{M}}_k({\bf x}_\theta, {\bf x}_{k+1})`
        """
        return self.F_list
        
    def assimilate(self, pi, ll, *args, **kwargs):
        r""" Assimilate one piece of data :math:`\left( \pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k, \Theta \right), \log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}, \Theta\right) \right)`.

        Given the new piece of data
        :math:`\left( \pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k, \Theta \right), \log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}, \Theta\right) \right)`,
        determine the maps pushing forward :math:`\mathcal{N}(0,{\bf I})`
        to the filtering/forecast distributions
        :math:`\{\pi\left(\Theta, {\bf Z}_k \middle\vert {\bf y}_{0:k} \right)\}_k`.
        
        Args:
          pi (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
            transition distribution
            :math:`\pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k, \Theta \right)`
          ll (:class:`LogLikelihood<TransportMaps.Functionals.LogLikelihood>`):
            log-likelihood
            :math:`\log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}, \Theta\right)`. The value ``None`` stands for missing observation.
          *args, **kwargs: arguments required by the particular sub-classes implementations
            of :func:`_assimilation_step`.

        .. note:: This method requires the implementation of the function
           :func:`_assimilation_step` in sub-classes
        """
        # Append transition and log-likelihood in self.pi
        self._pi.append(pi, ll)
        self._nsteps += 1
        # Approximation
        self._assimilation_step(*args, **kwargs)

    def _assimilation_step(self, *args, **kwargs):
        r""" [Abstract] Implements the map approximation for one step in the sequential inference.

        .. raises:: NotImplementedError
        """
        raise NotImplementedError("To be implemented in sub-classes.")

    def get_filtering_map_list(self):
        r"""
        .. deprecated:: Use :attr:`filtering_map_list` instead
        """
        deprecate("filtering_map_list", "1.0b3",
                  "Use property filtering_map_list instead")
        return self.filtering_map_list

class Smoother(Filter):
    r""" Perform the on-line smoothing and filtering of a sequential Hidded Markov chain.

    Given the prior distribution on the hyper-parameters :math:`\pi(\Theta)`,
    provides the functions neccessary to assimilate new pieces of data or
    missing data 
    (defined in terms of transition densities
    :math:`\pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k, \Theta \right)`
    and log-likelihoods
    :math:`\log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}, \Theta\right)`),
    to return the map pushing forward :math:`\mathcal{N}(0,{\bf I})`
    to the smoothing distribution
    :math:`\pi\left(\Theta, {\bf Z}_\Lambda \middle\vert {\bf y}_\Xi \right)`
    and to return the maps pushing forward :math:`\mathcal{N}(0,{\bf I})`
    to the filtering/forecast distributions
    :math:`\{\pi\left(\Theta, {\bf Z}_k \middle\vert {\bf y}_{0:k} \right)\}_k`.

    For more details see also :cite:`Spantini2017` and the
    `tutorial <example-sequential-stocvol-6d.html>`_.

    Args:
      pi_hyper (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        prior distribution on the hyper-parameters :math:`\pi(\Theta)`

    .. note:: This is a super-class. Part of its methods need to be implemented
      by sub-classes.
    """
    def __init__(self, pi_hyper=None):
        super(Smoother,self).__init__(pi_hyper)
        self.L_list = []

    @property
    def smoothing_map(self):
        r""" Returns the map :math:`\mathfrak{T}` pushing forward :math:`\mathcal{N}(0,{\bf I})` to the smoothing distribution :math:`\pi\left(\Theta, {\bf Z}_\Lambda \middle\vert {\bf y}_\Xi\right)`.

        The map :math:`\mathfrak{T}` is given by the composition
        :math:`T_0 \circ \cdots \circ T_{k-1}` maps constructed in
        :math:`k` assimilation steps.

        Returns:
          (:class:`TransportMap<TransportMaps.Maps.TransportMap>`) -- the map :math:`\mathfrak{T}`
        """
        if self.nsteps >= 0:
            return ListCompositeTransportMap( self.L_list )
        else:
            raise RuntimeError("No step assimilated yet!")
        
    def assimilate(self, pi, ll, *args, **kwargs):
        r""" Assimilate one piece of data :math:`\left( \pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k, \Theta \right), \log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}, \Theta\right) \right)`.

        Given the new piece of data
        :math:`\left( \pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k, \Theta \right), \log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}, \Theta\right) \right)`,
        retrieve the :math:`k`-th Markov component :math:`\pi^k` of :math:`\pi`,
        determine the transport map

        .. math::

           \mathfrak{M}_k({\boldsymbol \theta}, {\bf z}_k, {\bf z}_{k+1}) = \left[
           \begin{array}{l}
           \mathfrak{M}^\Theta_k({\boldsymbol \theta}) \\
           \mathfrak{M}^0_k({\boldsymbol \theta}, {\bf z}_k, {\bf z}_{k+1}) \\
           \mathfrak{M}^1_k({\boldsymbol \theta}, {\bf z}_{k+1})
           \end{array}
           \right] = Q \circ R_k \circ Q

        that pushes forward :math:`\mathcal{N}(0,{\bf I})` to :math:`\pi^k`, and
        embed it into the linear map which will remove the desired conditional
        dependencies from :math:`\pi`.
        
        Args:
          pi (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
            transition distribution
            :math:`\pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k, \Theta \right)`
          ll (:class:`LogLikelihood<TransportMaps.Functionals.LogLikelihood>`):
            log-likelihood
            :math:`\log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}, \Theta\right)`. The value ``None`` stands for missing observation.
          *args, **kwargs: arguments required by the particular sub-classes implementations
            of :func:`_assimilation_step`.

        .. note:: This method requires the implementation of the function
           :func:`_assimilation_step` in sub-classes
        """
        super(Smoother, self).assimilate(pi, ll, *args, **kwargs)

    def get_smoothing_map(self):
        r"""
        .. deprecated:: Use :attr:`filtering_map_list` instead
        """
        deprecate("get_smoothing_map", "1.0b3",
                  "Use property smoothing_map instead")
        return self.smoothing_map


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import unittest
import numpy as np
import numpy.random as npr
import numpy.linalg as npla
from scipy import stats

import TransportMaps as TM
import TransportMaps.Likelihoods as LKL
import TransportMaps.Maps as MAPS
import TransportMaps.Distributions as DIST
import TransportMaps.Distributions.Inference as INFDIST
import TransportMaps.Distributions.Decomposable as DECDIST
import TransportMaps.Distributions.Examples as DIST_EX
import TransportMaps.Distributions.Examples.StochasticVolatility as SV
from TransportMaps import FiniteDifference as FD

try:
    import mpi_map
    MPI_SUPPORT = True
except:
    MPI_SUPPORT = False

class DistributionTest(unittest.TestCase):

    def setUp(self):
        npr.seed(5)
        # Defined in subclasses
        self.d = self.get_d()
        self.distribution = self.get_distribution()
        self.sampling_d = self.get_sampling_distribution()
        self.integ = self.get_analytic_integral()
        self.f = self.get_test_function()
        self.mc_eps = self.get_mc_eps()
        self.quad_eps = self.get_quad_eps()

    def get_d(self):
        raise NotImplementedError("Define the dimension in sub-class")

    def get_sampling_distribution(self):
        return self.distribution # Default is the same

    def get_distribution(self):
        raise NotImplementedError("Define the distribution in sub-class")

    def get_analytic_integral(self):
        r""" Value :math:`\mathbb{E}_\pi[x]`
        """
        raise NotImplementedError("Define the analytic integral with respect " +
                                  "to the distribution sub-class")

    def get_test_function(self):
        raise NotImplementedError("Define the test function in sub-class")

    def get_mc_eps(self):
        raise NotImplementedError("Define eps for mc test in sub-class")

    def get_quad_eps(self):
        raise NotImplementedError("Define eps for quadrature tests in sub-class")

    def test_rvs(self):
        n = 100
        ninc = 5
        success = False
        nit = 0
        while not success and nit < ninc:
            n *= 10
            x = self.sampling_d.rvs(n)
            intapp = np.sum(self.f(x))/float(n)
            success = np.isclose(intapp,self.integ,atol=self.mc_eps)
            nit += 1
        self.assertTrue(success)

    def test_log_pdf(self):
        n = 100
        x = self.sampling_d.rvs(n)
        pdf = self.distribution.pdf(x)
        log_pdf = self.distribution.log_pdf(x)
        self.assertTrue( np.allclose( np.log(pdf), log_pdf ) )

    def test_grad_x_log_pdf(self):
        n = 100
        dx = 1e-5
        x = self.sampling_d.rvs(n)
        gxlp_exa = self.distribution.grad_x_log_pdf(x)
        gxlp_fd = FD.grad_x_fd(self.distribution.log_pdf, x, dx, None)
        self.assertTrue( np.allclose(gxlp_fd, gxlp_exa, rtol=50*dx, atol=10*dx) )

    def test_hess_x_log_pdf(self):
        n = 100
        dx = 1e-5
        x = self.sampling_d.rvs(n) / 5. # Closer to 0 finite difference works better.
        hxlp_exa = self.distribution.hess_x_log_pdf(x)
        hxlp_fd = FD.grad_x_fd(self.distribution.grad_x_log_pdf, x, dx, None)
        self.assertTrue( np.allclose(hxlp_fd, hxlp_exa, rtol=50*dx, atol=10*dx) )

    def test_mean_log_pdf(self):
        n = 100
        ninc = 5
        success = False
        nit = 0
        while not success and nit < ninc:
            n *= 10
            x = self.sampling_d.rvs(n)
            samps = self.distribution.log_pdf(x)
            intapp = np.sum(samps)/float(n)
            success = np.isclose(intapp, self.distribution.mean_log_pdf(),
                                 atol=self.mc_eps)
            nit += 1
        self.assertTrue(success)

class ConditionalDistributionTest(unittest.TestCase):

    def setUp(self):
        npr.seed(5)
        # Defined in subclasses
        self.distribution = self.get_distribution()
        self.dx = self.distribution.dim
        self.dy = self.distribution.dim_y
        self.d = self.get_d()
        self.sampling_d = self.get_sampling_distribution()

    def get_d(self):
        return self.dx + self.dy

    def get_sampling_distribution(self):
        return DIST.StandardNormalDistribution(self.get_d())

    def get_distribution(self):
        raise NotImplementedError("Define the distribution in sub-class")

    def test_log_pdf(self):
        n = 100
        x = self.sampling_d.rvs(n)
        xin = x[:,:self.dx]
        yin = x[:,self.dx:]
        pdf = self.distribution.pdf(xin, yin)
        log_pdf = self.distribution.log_pdf(xin, yin)
        self.assertTrue( np.allclose( np.log(pdf), log_pdf ) )

    def test_grad_x_log_pdf(self):
        n = 100
        dx = 1e-5
        x = self.sampling_d.rvs(n)
        xin = x[:,:self.dx]
        yin = x[:,self.dx:]
        gxlp_exa = self.distribution.grad_x_log_pdf(xin, yin)
        def f(x, *args):
            xin = x[:,:self.dx]
            yin = x[:,self.dx:]
            return self.distribution.log_pdf(xin, yin)
        gxlp_fd = FD.grad_x_fd(f, x, dx, None)
        self.assertTrue( np.allclose(gxlp_fd, gxlp_exa, rtol=50*dx, atol=10*dx) )

    def test_hess_x_log_pdf(self):
        n = 100
        dx = 1e-5
        x = self.sampling_d.rvs(n) / 5. # Closer to 0 finite difference works better.
        xin = x[:,:self.dx]
        yin = x[:,self.dx:]
        hxlp_exa = self.distribution.hess_x_log_pdf(xin, yin)
        def f(x, *args):
            xin = x[:,:self.dx]
            yin = x[:,self.dx:]
            return self.distribution.grad_x_log_pdf(xin, yin)
        hxlp_fd = FD.grad_x_fd(f, x, dx, None)
        self.assertTrue( np.allclose(hxlp_fd, hxlp_exa, rtol=50*dx, atol=10*dx) )
        
class GaussianDistributionSigmaTests(DistributionTest):

    def get_distribution(self):
        mu = stats.norm().rvs(self.d)
        S = stats.norm().rvs(self.d**2).reshape((self.d,self.d))
        sig = np.dot(S.T, S)
        return DIST.GaussianDistribution(mu, sig)    

    def get_d(self):
        return 5

    def get_test_function(self):
        f = lambda x: np.sum(x, axis=0)
        return f

    def get_analytic_integral(self):
        return np.sum(self.distribution.mu)

    def get_mc_eps(self):
        return 1e-2

    def get_quad_eps(self):
        return 1e-10

    def test_gauss_quadrature(self):
        order = 1
        success = 0
        maxord = 20
        while not success and order <= maxord:
            (x,w) = self.distribution.quadrature(3, [order]*self.d)
            intapp = np.dot(self.f(x.T), w)
            success = np.isclose(intapp,self.integ,atol=self.quad_eps)
            order += 1
        self.assertTrue(success)

class GaussianDistributionPrecisionTests(DistributionTest):

    def get_distribution(self):
        mu = stats.norm().rvs(self.d)
        S = stats.norm().rvs(self.d**2).reshape((self.d,self.d))
        sig = np.dot(S.T, S)
        prec = npla.solve(sig, np.eye(self.d))
        return DIST.GaussianDistribution(mu, precision=prec)    

    def get_d(self):
        return 5

    def get_test_function(self):
        f = lambda x: np.sum(x, axis=0)
        return f

    def get_analytic_integral(self):
        return np.sum(self.distribution.mu)

    def get_mc_eps(self):
        return 1e-2

    def get_quad_eps(self):
        return 1e-10

    def test_gauss_quadrature(self):
        order = 1
        success = 0
        maxord = 20
        while not success and order <= maxord:
            (x,w) = self.distribution.quadrature(3, [order]*self.d)
            intapp = np.dot(self.f(x.T), w)
            success = np.isclose(intapp,self.integ,atol=self.quad_eps)
            order += 1
        self.assertTrue(success)


class StandardNormalDistributionTest(GaussianDistributionSigmaTests):

    def get_distribution(self):
        return DIST.StandardNormalDistribution(self.d)

class LogNormalDistributionTest(DistributionTest):
    def get_distribution(self):
        return DIST.LogNormalDistribution(.3,0.,2.)
    def get_d(self): return 1
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class LogisticDistributionTest(DistributionTest):
    def get_distribution(self):
        return DIST.LogisticDistribution(1.,1.5)
    def get_d(self): return 1
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class GammaDistributionTest(DistributionTest):
    def get_distribution(self):
        return DIST.GammaDistribution(2.,5.)
    def get_d(self): return 1
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class BetaDistributionTest(DistributionTest):
    def get_distribution(self):
        return DIST.BetaDistribution(2.,5.)
    def get_d(self): return 1
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class GumbelDistributionTest(DistributionTest):
    def get_distribution(self):
        return DIST.GumbelDistribution(2.,5.)
    def get_d(self): return 1
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class WeibullDistributionTest(DistributionTest):
    def get_distribution(self):
        return DIST.WeibullDistribution(2.,2.,2.)
    def get_d(self): return 1
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class BananaDistributionTest(DistributionTest):
    def get_distribution(self):
        return DIST.BananaDistribution(a=2.,b=2.,mu=np.zeros(2),sigma2=np.eye(2))
    def get_d(self): return 2
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class FactorizedDistributionTest(DistributionTest):
    r"""
    We test the joint distribution :math:`\pi(x_0,x_1)=\pi_1(x_1|x_0)\pi_2(x_0)` defined by

    .. math::

       x_0 \sim \mathcal{N}(0,1) \\
       x_1 | x_0 \sim \mathcal{N}(x_0^2, 1)
    
    """
    def get_sampling_distribution(self):
        return DIST.StandardNormalDistribution(self.d)
    def get_distribution(self):
        return DIST_EX.FactorizedBananaDistribution()
    def get_d(self): return 2
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class LinearBayesPosteriorDistributionTest(DistributionTest):
    r"""
    Test the Bayesian posterior for a linear Gaussian additive model with
    Gaussian prior.

    .. math::

       {\bf y} = {\bf c} + {\bf T}{\bf x} + \varepsilon \;, \quad
       \varepsilon \sim \mathcal{N}(\mu, \Sigma) \;, \quad
       {\bf x} \sim \mathcal{N}(\mu_x, \Sigma_x)

    To make it more interesting we set :math:`d_y=2` and :math:`d_x=3`
    """
    def get_sampling_distribution(self):
        return DIST.StandardNormalDistribution(self.d)
    def get_distribution(self):
        dx = self.d
        dy = dx - 1
        # Prior
        mux = npr.randn(dx)
        sigmax = npr.randn(dx**2).reshape((dx,dx))
        sigmax = np.dot(sigmax, sigmax.T)
        prior = DIST.GaussianDistribution(mux, sigmax)
        # Likelihood
        mu = npr.randn(dy)
        sigma = npr.randn(dy**2).reshape((dy,dy))
        sigma = np.dot(sigma, sigma.T)
        c = npr.randn(dy)
        T = npr.randn(dy*dx).reshape((dy,dx))
        y = c + np.dot(T, prior.rvs(1)[0,:])
        logL = LKL.AdditiveLinearGaussianLogLikelihood(y, c, T, mu, sigma)
        # Posterior
        pi = INFDIST.BayesPosteriorDistribution(logL, prior)
        return pi
    def get_d(self): return 3
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class NonLinearBayesPosteriorDistributionTest(DistributionTest):
    r"""
    Test the Bayesian posterior for a non-linear Gaussian additive model with
    Gaussian prior.

    .. math::

       {\bf y} = T({\bf x}) + \varepsilon \;, \quad
       \varepsilon \sim \mathcal{N}(\mu, \Sigma) \;, \quad
       {\bf x} \sim \mathcal{N}(\mu_x, \Sigma_x)

    To make it more interesting we set :math:`d_y=2` and :math:`d_x=3`
    """
    def get_sampling_distribution(self):
        return DIST.StandardNormalDistribution(self.d)
    def get_distribution(self):
        dx = self.d
        dy = dx - 1
        # Prior
        mux = npr.randn(dx)
        sigmax = npr.randn(dx**2).reshape((dx,dx))
        sigmax = np.dot(sigmax, sigmax.T)
        prior = DIST.GaussianDistribution(mux, sigmax)
        # Likelihood
        mu = npr.randn(dy)
        sigma = npr.randn(dy**2).reshape((dy,dy))
        sigma = np.dot(sigma, sigma.T)
        obs_pi = DIST.GaussianDistribution(mu, sigma)
        Tfull = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(dx, 2)
        coeffs = Tfull.get_identity_coeffs()
        coeffs += npr.randn(len(coeffs)) / 10.
        Tfull.coeffs = coeffs
        T = MAPS.TransportMap( Tfull.active_vars[dx-dy:],
                               Tfull.approx_list[dx-dy:] )
        y = T.evaluate(prior.rvs(1)).flatten()
        logL = LKL.AdditiveLogLikelihood(y, obs_pi, T)
        # Posterior
        pi = INFDIST.BayesPosteriorDistribution(logL, prior)
        return pi
    def get_d(self): return 3
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class AR1TransitionDistributionTest(ConditionalDistributionTest):
    r"""
    We test for the transition probability associated to the folllowing process

    .. math::

       {\bf Z}_{k+1} = T({\bf Z}_k, \Theta) + \varepsilon
       \;, \quad
       \varepsilon \sim \nu_\pi

    where :math:`T` is a non-linear map and \nu_\pi is the banana distribution.
    """
    def get_distribution(self):
        # \nu_\pi
        a = 1.
        b = 1.
        mu = np.zeros(2)
        sigma2 = np.array([[1., 0.9],[0.9, 1.]])
        nupi = DIST.BananaDistribution(a, b, mu, sigma2)
        # T:R^(2+2) -> R^2 (bottom part of a lower triangular linear span map)
        Tfull = TM.Default_IsotropicLinearSpanTriangularTransportMap(4, 2)
        T = MAPS.TransportMap( Tfull.active_vars[2:],
                               Tfull.approx_list[2:] )
        # Build transition probability
        ar1trans = DECDIST.AR1TransitionDistribution(nupi, T)
        return ar1trans
    
class StochasticVolatilityDistributionTest(DistributionTest):
    r"""
    We test on the joint distribution of the stochastic volatility with hyper-parameters and
    4 steps.
    """
    def get_sampling_distribution(self):
        return DIST.StandardNormalDistribution(self.d)
    def get_distribution(self):
        # PRIOR HYPER PARAMTERS
        is_mu_h = True
        is_sigma_h = True
        is_phi_h = True
        pi_hyper = SV.PriorHyperParameters(is_mu_h, is_sigma_h, is_phi_h, 1)
        # INITIAL AND TRANSITION PROBABILITIES
        mu_h = SV.IdentityFunction()
        phi_h = SV.F_phi(3.,1.)
        sigma_h = SV.ConstantFunction(0.25)
        pi_ic = SV.PriorDynamicsInitialConditions(
            is_mu_h, mu_h, is_sigma_h, sigma_h, is_phi_h, phi_h)
        pi_trans = SV.PriorDynamicsTransition(
            is_mu_h, mu_h, is_sigma_h, sigma_h, is_phi_h, phi_h)
        # GENERATE OBSERVATIONS
        mu = -.5
        phi = .95
        sigma = 0.25
        nsteps = 4
        yB, ZA = SV.generate_data(nsteps, mu, sigma, phi)
        yB[2] = None # Missing data y2
        # DEFINE POSTERIOR AND ASSIMILATE DATA
        pi = DECDIST.SequentialHiddenMarkovChainDistribution([], [], pi_hyper)
        for n, yt in enumerate(yB):
            if yt is None:
                ll = None
            else:
                ll = SV.LogLikelihood(yt, is_mu_h, is_sigma_h, is_phi_h)
            if n == 0: pin = pi_ic
            else: pin = pi_trans
            pi.append(pin, ll)
        return pi
    def get_d(self): return 7
    def get_analytic_integral(self): return None
    def get_test_function(self): return None
    def get_mc_eps(self): return None
    def get_quad_eps(self): return None
    @unittest.skip("Not defined")
    def test_rvs(self): pass
    @unittest.skip("Not defined")
    def test_mean_log_pdf(self): pass

class MarkovComponentZeroStochasticVolatilityDistributionTest(
        StochasticVolatilityDistributionTest):
    r"""
    We test on the joint distribution of the stochastic volatility with hyper-parameters and
    4 steps.
    """
    def get_distribution(self):
        pi = super(MarkovComponentZeroStochasticVolatilityDistributionTest,
                   self).get_distribution()
        mkv = pi.get_MarkovComponent(0, n=2)
        return mkv
    def get_d(self): return 6

class MarkovComponentOneStochasticVolatilityDistributionTest(
        StochasticVolatilityDistributionTest):
    r"""
    We test on the joint distribution of the stochastic volatility with hyper-parameters and
    4 steps.
    """
    def get_distribution(self):
        pi = super(MarkovComponentOneStochasticVolatilityDistributionTest,
                   self).get_distribution()
        # Construct random map
        hdim = pi.hyper_dim
        sdim = pi.state_dim
        T = TM.Default_IsotropicIntegratedSquaredTriangularTransportMap(
            sdim+hdim, 1)
        coeffs = T.get_identity_coeffs()
        coeffs += npr.randn( len(coeffs) ) / 10.
        T.coeffs = coeffs
        # Retriever hyper map and Xt map
        H = MAPS.TransportMap( T.active_vars[:hdim],
                               T.approx_list[:hdim] )
        Mkm1 = MAPS.TransportMap( T.active_vars[hdim:hdim+sdim],
                                  T.approx_list[hdim:hdim+sdim] )
        mkv = pi.get_MarkovComponent(1, n=2, state_map=Mkm1, hyper_map=H)
        return mkv
    def get_d(self): return 6

    
def build_suite(ttype='all'):
    suite_stdnorm = unittest.TestLoader().loadTestsFromTestCase( StandardNormalDistributionTest )
    suite_gaussian_sigma = unittest.TestLoader().loadTestsFromTestCase( GaussianDistributionSigmaTests )
    suite_gaussian_precision = unittest.TestLoader().loadTestsFromTestCase( GaussianDistributionPrecisionTests )
    suite_lognormal = unittest.TestLoader().loadTestsFromTestCase( LogNormalDistributionTest )
    suite_logistic = unittest.TestLoader().loadTestsFromTestCase( LogisticDistributionTest )
    suite_gamma = unittest.TestLoader().loadTestsFromTestCase( GammaDistributionTest )
    suite_beta = unittest.TestLoader().loadTestsFromTestCase( BetaDistributionTest )
    suite_gumbel = unittest.TestLoader().loadTestsFromTestCase( GumbelDistributionTest )
    suite_weibull = unittest.TestLoader().loadTestsFromTestCase( WeibullDistributionTest )
    suite_banana = unittest.TestLoader().loadTestsFromTestCase( BananaDistributionTest )
    suite_factorized = unittest.TestLoader().loadTestsFromTestCase(
        FactorizedDistributionTest )
    suite_bayeslinear = unittest.TestLoader().loadTestsFromTestCase(
        LinearBayesPosteriorDistributionTest )
    suite_bayesnonlin = unittest.TestLoader().loadTestsFromTestCase(
        NonLinearBayesPosteriorDistributionTest )
    suite_ar1trans = unittest.TestLoader().loadTestsFromTestCase(
        AR1TransitionDistributionTest )
    suite_stocvol = unittest.TestLoader().loadTestsFromTestCase(
        StochasticVolatilityDistributionTest )
    suite_m0stocvol = unittest.TestLoader().loadTestsFromTestCase(
        MarkovComponentZeroStochasticVolatilityDistributionTest )
    suite_m1stocvol = unittest.TestLoader().loadTestsFromTestCase(
        MarkovComponentOneStochasticVolatilityDistributionTest )
    # Group suites
    suites_list = []
    if ttype in ['all', 'serial']:
        suites_list += [ suite_stdnorm, suite_gaussian_sigma,
                         suite_gaussian_precision,
                         suite_lognormal, suite_logistic, suite_gamma, suite_beta,
                         suite_gumbel, suite_weibull, suite_banana,
                         suite_factorized, suite_bayeslinear, suite_bayesnonlin,
                         suite_ar1trans,
                         suite_stocvol, suite_m0stocvol, suite_m1stocvol]
    all_suites = unittest.TestSuite( suites_list )
    return all_suites

def run_tests(ttype='all'):
    all_suites = build_suite(ttype)
    # RUN
    unittest.TextTestRunner(verbosity=2).run(all_suites)

if __name__ == '__main__':
    run_tests()



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2016 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Daniele Bigoni
# E-mail: dabi@limitcycle.it
#

import unittest
import numpy as np
import numpy.random as npr
import numpy.linalg as npla

try:
    import mpi_map
    MPI_SUPPORT = True
except:
    MPI_SUPPORT = False

class Functionals_DerivativeChecks(unittest.TestCase):

    def setUp(self):
        npr.seed(1)
        
        import TransportMaps.Densities as DENS
        self.dim = 2
        self.order = 3
        self.fd_eps = 1e-5
        self.qtype = 3
        self.qpar = [5]*self.dim
        self.density = DENS.StandardNormalDensity(self.dim)
        (x,w) = self.density.quadrature(self.qtype, self.qpar)
        self.x = x
        self.w = w
        self.build_tm_approx()

    def test_grad_a(self):
        import TransportMaps.FiniteDifference as FD

        approx = self.approx
        coeffs = self.coeffs
        x = self.x

        # Define transport map, gradient
        def f(a, *args, **kwargs):
            approx.coeffs = a
            out = approx.evaluate(x)
            return out
        def grad_a_f(a, *args, **kwargs):
            approx.coeffs = a
            out = approx.grad_a(x)
            return out

        # Check gradient transport map
        flag = FD.check_grad_a(f, grad_a_f, coeffs, self.fd_eps,
                               None, verbose=False)
        self.assertTrue( flag )

    def test_grad_x_partial_xd(self):
        import TransportMaps.FiniteDifference as FD

        approx = self.approx
        coeffs = self.coeffs
        x = self.x

        # Define transport map, gradient
        def f(x, *args, **kwargs):
            out = approx.partial_xd(x, {})
            return out
        def der(x, *args, **kwargs):
            out = approx.grad_x_partial_xd(x, {})
            return out

        # Check gradient transport map
        flag = FD.check_grad_x(f, der, x, self.fd_eps,
                               None, verbose=False)
        self.assertTrue( flag )

    def test_hess_x_partial_xd(self):
        import TransportMaps.FiniteDifference as FD

        approx = self.approx
        coeffs = self.coeffs
        x = self.x

        # Define transport map, gradient
        def f(x, *args, **kwargs):
            out = approx.grad_x_partial_xd(x, {})
            return out
        def der(x, *args, **kwargs):
            out = approx.hess_x_partial_xd(x, {})
            return out

        # Check gradient transport map
        flag = FD.check_grad_x(f, der, x, self.fd_eps,
                               None, verbose=False)
        self.assertTrue( flag )

    def test_grad_a_grad_x_partial_xd(self):
        import TransportMaps.FiniteDifference as FD

        approx = self.approx
        coeffs = self.coeffs
        x = self.x

        # Define transport map, gradient
        def f(a, *args, **kwargs):
            approx.coeffs = a
            out = approx.grad_x_partial_xd(x, {})
            return out
        def grad_a_f(a, *args, **kwargs):
            approx.coeffs = a
            out = approx.grad_a_grad_x_partial_xd(x, {})
            return np.transpose(out,(0,2,1))

        # Check gradient transport map
        flag = FD.check_grad_a(f, grad_a_f, coeffs, self.fd_eps,
                               None, verbose=False)
        self.assertTrue( flag )
        
    def test_grad_a_hess_x_partial_xd(self):
        import TransportMaps.FiniteDifference as FD

        approx = self.approx
        coeffs = self.coeffs
        x = self.x

        # Define transport map, gradient
        def f(a, *args, **kwargs):
            approx.coeffs = a
            out = approx.hess_x_partial_xd(x, {})
            return out
        def grad_a_f(a, *args, **kwargs):
            approx.coeffs = a
            out = approx.grad_a_hess_x_partial_xd(x, {})
            return np.transpose(out,(0,2,3,1))

        # Check gradient transport map
        flag = FD.check_grad_a(f, grad_a_f, coeffs, self.fd_eps,
                               None, verbose=False)
        self.assertTrue( flag )
        
class LinearSpanFunctional( object ):
    """ Linear span approximation
    """
    def build_tm_approx(self):
        import numpy.random as npr
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        basis_list = [S1D.HermiteProbabilistsPolynomial()] * (self.dim)
        orders_list = [self.order] * (self.dim)
        self.approx = FUNC.MonotonicLinearSpanApproximation(basis_list, spantype="full",
                                                                order_list=orders_list)
        # Set coefficients for linear map
        self.coeffs = npr.randn(self.approx.n_coeffs)/10.

class MonotonicIntegratedExponentialFunctional( object ):
    """ Linear span approximation
    """
    def build_tm_approx(self):
        import numpy.random as npr
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps as TM
        import TransportMaps.Functionals as FUNC
        basis_list = [S1D.HermiteProbabilistsPolynomial()] * (self.dim)
        orders_list = [self.order] * (self.dim - 1) + [0]
        constant = FUNC.LinearSpanApproximation(basis_list, spantype="full",
                                                                order_list=orders_list)
        basis_list = [S1D.ConstantExtendedHermiteProbabilistsFunction()] * (self.dim)
        orders_list = [self.order] * (self.dim)
        exponential = FUNC.LinearSpanApproximation(basis_list, spantype="full",
                                                                order_list=orders_list)
        self.approx = FUNC.MonotonicIntegratedExponentialApproximation(constant, exponential)
        # Set coefficients for linear map
        self.coeffs = npr.randn(self.approx.n_coeffs)/10.

# FULL ORDER LINEAR SPAN
class OnTheFly_LinearSpanFunctional_DerivativeChecks( LinearSpanFunctional,
                                                      Functionals_DerivativeChecks ):
    pass

class OnTheFly_MonotonicIntegratedExponentialFunctional_DerivativeChecks( MonotonicIntegratedExponentialFunctional,
                                                      Functionals_DerivativeChecks ):
    pass

def build_suite(ttype='all'):
    # Full order linear span
    suite_of_lsf_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_LinearSpanFunctional_DerivativeChecks )
    suite_of_mief_dc = unittest.TestLoader().loadTestsFromTestCase(
        OnTheFly_MonotonicIntegratedExponentialFunctional_DerivativeChecks )
    # GROUP SUITES
    suites_list = []
    if ttype in ['all','serial']:
        suites_list += [ suite_of_lsf_dc, suite_of_mief_dc ]
    all_suites = unittest.TestSuite( suites_list )
    return all_suites

def run_tests(ttype='all'):
    all_suites = build_suite(ttype)
    # RUN
    unittest.TextTestRunner(verbosity=2).run(all_suites)

if __name__ == '__main__':
    run_tests()



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import warnings
import numpy as np
import numpy.linalg as npla
import scipy.stats as stats
import scipy.linalg as scila

import SpectralToolbox.Spectral1D as S1D
import SpectralToolbox.SpectralND as SND

from TransportMaps.Distributions.DistributionBase import *
from TransportMaps.Distributions.TransportMapDistributions import *

__all__ = ['FrozenDistribution_1d',
           'GaussianDistribution', 'StandardNormalDistribution',
           'LogNormalDistribution', 'LogisticDistribution',
           'GammaDistribution', 'BetaDistribution',
           'WeibullDistribution',
           'GumbelDistribution', 'BananaDistribution']

nax = np.newaxis

class FrozenDistribution_1d(Distribution):
    r""" [Abstract] Generic frozen distribution 1d
    """
    def __init__(self):
        super(FrozenDistribution_1d,self).__init__(1)
        self.base = StandardNormalDistribution(1)
        self.scipy_base = stats.norm()
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    def quadrature(self, qtype, qparams, *args, **kwargs):
        r""" Generate quadrature points and weights.

        Types of quadratures:

        Monte-Carlo (``qtype==0``)
           ``qparams``: (:class:`int`) -- number of samples

        Quasi-Monte-Carlo (``qtype==1``)
           ``qparams``: (:class:`int`) -- number of samples

        Latin-Hypercube-Sampling (``qtype==2``)
           ``qparams``: (:class:`int`) -- number of samples

        Gauss-quadrature (``qtype==3``)
           ``qparams``: (:class:`list<list>` [:math:`d`]) -- orders for
           each dimension

        .. seealso:: :func:`Distribution.quadrature`
        """
        if qtype == 0:
            # Monte Carlo sampling
            x = self.rvs(qparams)
            w = np.ones(qparams)/float(qparams)
        elif qtype == 1:
            # Quasi-Monte Carlo sampling
            raise NotImplementedError("Not implemented")
        elif qtype == 2:
            # Latin-Hyper cube sampling
            raise NotImplementedError("Todo")
        elif qtype == 3:
            # Gaussian quadrature
            (x1,w) = self.base.quadrature(3, qparams)
            x = self.dist.ppf( self.scipy_base.cdf(x1[:,0]) )[:,nax]
        else:
            raise ValueError("Quadrature type not recognized")
        return (x,w)

#########################################################
# Definitions of Gaussian and Standard normal densities #
#########################################################
class GaussianDistribution(Distribution):
    r""" Multivariate Gaussian distribution :math:`\pi`

    Args:
      mu (:class:`ndarray<numpy.ndarray>` [:math:`d`]): mean vector
      sigma (:class:`ndarray<numpy.ndarray>` [:math:`d,d`]): covariance matrix
      precision (:class:`ndarray<numpy.ndarray>` [:math:`d,d`]): precision matrix
    """

    def __init__(self, mu, sigma=None, precision=None):
        if (sigma is not None) and (precision is not None):
            raise ValueError("The fields sigma and precision are mutually " +
                             "exclusive")
        super(GaussianDistribution,self).__init__(mu.shape[0])
        self._mu = None
        self._sigma = None
        self._precision = None
        self.mu = mu
        if sigma is not None:
            self.sigma = sigma
        if precision is not None:
            self.precision = precision

    @property
    def mu(self):
        return self._mu

    @mu.setter
    def mu(self, mu):
        if (self._sigma is not None and mu.shape[0] != self._sigma.shape[0]) or \
           (self._precision is not None and mu.shape[0] != self._precision.shape[0]):
            raise ValueError("Dimension d of mu and sigma/precision must be the same")
        self._mu = mu

    @property
    def sigma(self):
        return self._sigma

    @sigma.setter
    def sigma(self, sigma):
        if self.mu.shape[0] != sigma.shape[0] or self.mu.shape[0] != sigma.shape[1]:
            raise ValueError("Dimension d of mu and sigma must be the same")
        if self._sigma is None or np.any(self._sigma != sigma):
            self._sigma = sigma
            try:
                chol = scila.cho_factor(self._sigma, True) # True: lower triangular
            except scila.LinAlgError:
                # Obtain the square root from svd
                u,s,v = scila.svd(self._sigma)
                self.log_det_sigma = np.sum(np.log(s))
                self.det_sigma = np.exp( self.log_det_sigma )
                self.sampling_mat = u * np.sqrt(s)[np.newaxis,:]
                self.inv_sigma = np.dot(u * (1./s)[np.newaxis,:], v.T)
            else:
                self.det_sigma = np.prod(np.diag(chol[0]))**2.
                self.log_det_sigma = 2. * np.sum( np.log( np.diag(chol[0]) ) )
                self.sampling_mat = np.tril(chol[0])
                self.inv_sigma = scila.cho_solve(chol, np.eye(self.dim))

    @property
    def precision(self):
        return self._precision

    @precision.setter
    def precision(self, precision):
        if self.mu.shape[0] != precision.shape[0] or self.mu.shape[0] != precision.shape[1]:
            raise ValueError("Dimension d of mu and precision must be the same")
        if self._precision is None or np.any(self.inv_sigma != precision):
            self._precision = precision
            self.inv_sigma = precision
            chol = scila.cho_factor(self.inv_sigma, False) # False: upper triangular
            self.sigma = scila.cho_solve(chol, np.eye(self.dim))
            self.det_sigma = 1. / np.prod(np.diag(chol[0]))**2.
            self.log_det_sigma = - 2. * np.sum( np.log( np.diag(chol[0]) ) )
            self.sampling_mat = scila.solve_triangular(np.triu(chol[0]),
                                                       np.eye(self.dim),
                                                       lower=False)
            
    def rvs(self, m, *args, **kwargs):
        r""" Generate :math:`m` samples from the distribution.

        Args:
          m (int): number of samples

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- samples

        .. seealso:: :func:`Distribution.rvs`
        """
        z = stats.norm().rvs(m*self.dim).reshape((m,self.dim))
        samples = self._mu + np.dot( self.sampling_mat, z.T ).T
        return samples

    def quadrature(self, qtype, qparams, *args, **kwargs):
        r""" Generate quadrature points and weights.

        Types of quadratures:

        Monte-Carlo (``qtype==0``)
           ``qparams``: (:class:`int`) -- number of samples

        Quasi-Monte-Carlo (``qtype==1``)
           ``qparams``: (:class:`int`) -- number of samples

        Latin-Hypercube-Sampling (``qtype==2``)
           ``qparams``: (:class:`int`) -- number of samples

        Gauss-quadrature (``qtype==3``)
           ``qparams``: (:class:`list<list>` [:math:`d`]) -- orders for
           each dimension

        .. seealso:: :func:`Distribution.quadrature`
        """
        if qtype == 0:
            # Monte Carlo sampling
            x = self.rvs(qparams)
            w = np.ones(qparams)/float(qparams)
        elif qtype == 1:
            # Quasi-Monte Carlo sampling
            raise NotImplementedError("Not implemented")
        elif qtype == 2:
            # Latin-Hyper cube sampling
            raise NotImplementedError("Todo")
        elif qtype == 3:
            # Gaussian quadrature
            # Generate first a standard normal quadrature
            # then apply the Cholesky transform
            P = SND.PolyND( [S1D.HermiteProbabilistsPolynomial()] * self.dim )
            (x,w) = P.GaussQuadrature(qparams, norm=True)
            x = np.dot( self.sampling_mat, x.T ).T
            x += self._mu
        else:
            raise ValueError("Quadrature type not recognized")
        return (x,w)

    def pdf(self, x, *args, **kwargs):
        r""" Evaluate :math:`\pi(x)`

        .. seealso:: :func:`Distribution.pdf`
        """
        return np.exp( self.log_pdf(x) )

    def log_pdf(self, x, *args, **kwargs):
        r""" Evaluate :math:`\log\pi(x)`

        .. seealso:: :func:`Distribution.log_pdf`
        """
        b = x - self._mu
        sol = np.dot( self.inv_sigma, b.T ).T
        out = - .5 * np.einsum('...i,...i->...', b, sol) \
              - self.dim * .5 * np.log(2.*np.pi) \
              - .5 * self.log_det_sigma
        return out.flatten()

    def grad_x_log_pdf(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x}\log\pi(x)`

        .. seealso:: :func:`Distribution.grad_x_log_pdf`
        """
        b = x - self._mu
        return - np.dot( self.inv_sigma, b.T ).T

    def hess_x_log_pdf(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf x}\log\pi(x)`

        .. seealso:: :func:`Distribution.hess_x_log_pdf`
        """
        return - np.ones(x.shape[0])[:,nax,nax] * self.inv_sigma[nax,:,:]

    def mean_log_pdf(self):
        r""" Evaluate :math:`\mathbb{E}_{\pi}[\log \pi]`.

        .. seealso:: :func:`Distribution.mean_log_pdf`
        """
        return - .5 * ( self.dim * np.log(2*np.pi) + self.dim + self.log_det_sigma )

class StandardNormalDistribution(GaussianDistribution, ProductDistribution):
    r""" Multivariate Standard Normal distribution :math:`\pi`.

    Args:
      d (int): dimension
    """

    def __init__(self, dim):
        self._mu = np.zeros(dim)
        self._sigma = np.eye(dim)
        self._precision = np.eye(dim)
        self.log_det_sigma = 0.
        self.det_sigma = 1.
        self.sampling_mat = np.eye(dim)
        self.inv_sigma = np.eye(dim)
        Distribution.__init__(self, dim)

    def rvs(self, m, *args, **kwargs):
        r""" Generate :math:`m` samples from the distribution.

        Args:
          m (int): number of samples

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- samples

        .. seealso:: :func:`Distribution.rvs`
        """
        return stats.norm().rvs(m*self.dim).reshape((m,self.dim))

    def get_component(self, avars):
        r""" Return the measure :math:`\nu_{a_1}\times\cdots\times\nu_{a_k} = \mathcal{N}(0,{\bf I}_k)`

        Args:
          avars (list): list of coordinates to extract from :math:`\nu`
        """
        return StandardNormalDistribution(len(avars))

###############################################################
# Definition of miscellaneous densities (No sampling defined) #
###############################################################
class LogNormalDistribution(FrozenDistribution_1d):
    def __init__(self, s, mu, scale):
        super(LogNormalDistribution,self).__init__()
        self.s = s
        self.mu = mu
        self.scale = scale
        self.dist = stats.lognorm(s=s,
                                  loc=mu,
                                  scale=scale)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf( x ).flatten()
    def grad_x_pdf(self, x, *args, **kwargs):
        s = self.s
        m = self.mu
        d = self.dist
        return - d.pdf(x) * ( 1./(x-m) + np.log(x-m)/(s**2.*(x-m)) )
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf( x ).flatten()
    def grad_x_log_pdf(self, x, *args, **kwargs):
        s = self.s
        m = self.mu
        sc = self.scale
        return - 1./(x-m) * (np.log((x-m)/sc)/s**2. + 1)
    def hess_x_log_pdf(self, x, *args, **kwargs):
        s = self.s
        m = self.mu
        sc = self.scale
        return (1./(x-m)**2. * ( (np.log((x-m)/sc) + s**2. - 1.)/s**2. ))[:,:,nax]

class LogisticDistribution(FrozenDistribution_1d):
    def __init__(self, mu, s):
        super(LogisticDistribution,self).__init__()
        self.mu = mu
        self.s = s
        self.dist = stats.logistic(loc=mu,scale=s)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    def log_pdf(self, x, *args, **kwargs):
        # Log pdf with modified asymptotic behavior
        out = np.zeros(x.shape)
        g20 = (x >= -20)
        l20 = (x < -20)
        out[g20] = self.dist.logpdf(x[g20]).flatten()
        out[l20] = (x[l20].flatten() - self.mu)/self.s
        return out.flatten()
    def grad_x_log_pdf(self, x, *args, **kwargs):
        mu = self.mu
        s = self.s
        out = np.zeros(x.shape)
        g20 = (x >= -20)
        l20 = (x < -20)
        g = np.exp(-(x[g20]-mu)/s)
        out[g20] = -1./s + 2./s * g/(1+g)
        out[l20] = 1./s
        return out
    def hess_x_log_pdf(self, x, *args, **kwargs):
        mu = self.mu
        s = self.s
        out = np.zeros(x.shape)
        g20 = (x >= -20)
        l20 = (x < -20)
        g = np.exp(-(x[g20]-mu)/s)
        out[g20] = (- 2./s**2. * g/(1+g)**2.)
        out[l20] = 0.
        return out[:,:,nax]
    # def nabla3_x_log_pdf(self, x, params=None):
    #     mu = self.mu
    #     s = self.s
    #     g = np.exp(-(x-mu)/s)
    #     return (2./s**3. * g*(1-g)/(1+g)**3.)[:,:,nax,nax]

class GammaDistribution(FrozenDistribution_1d):
    def __init__(self, kappa, theta):
        super(GammaDistribution,self).__init__()
        self.kappa = kappa
        self.theta = theta
        self.dist = stats.gamma(kappa, scale=theta)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf(x).flatten()
    def grad_x_log_pdf(self, x, *args, **kwargs):
        k = self.kappa
        t = self.theta
        return (k-1.)/x - 1/t
    def hess_x_log_pdf(self, x, *args, **kwargs):
        k = self.kappa
        return ((1.-k)/x**2.)[:,:,nax]
    def nabla3_x_log_pdf(self, x, *args, **kwargs):
        k = self.kappa
        return (2.*(k-1.)/x**3.)[:,:,nax,nax]

class BetaDistribution(FrozenDistribution_1d):
    def __init__(self, alpha, beta):
        super(BetaDistribution,self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.dist = stats.beta(alpha, beta)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf(x).flatten()
    def grad_x_log_pdf(self, x, *args, **kwargs):
        a = self.alpha
        b = self.beta
        return (a-1.)/x + (b-1.)/(x-1.)
    def hess_x_log_pdf(self, x, *args, **kwargs):
        a = self.alpha
        b = self.beta
        out = (1.-a)/x**2. + (1-b)/(x-1.)**2.
        return out[:,:,nax]
    def nabla3_x_log_pdf(self, x, *args, **kwargs):
        a = self.alpha
        b = self.beta
        out = 2.*(a-1.)/x**3. + 2.*(b-1.)/(x-1.)**3.
        return out[:,:,nax,nax]

class GumbelDistribution(FrozenDistribution_1d):
    def __init__(self, mu, beta):
        super(GumbelDistribution,self).__init__()
        self.mu = mu
        self.beta = beta
        self.dist = stats.gumbel_r(loc=mu, scale=beta)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf(x).flatten()
    def grad_x_log_pdf(self, x, *args, **kwargs):
        m = self.mu
        b = self.beta
        z = (x-m)/b
        return (np.exp(-z)-1.)/b
    def hess_x_log_pdf(self, x, *args, **kwargs):
        m = self.mu
        b = self.beta
        z = (x-m)/b
        return (- np.exp(-z)/b**2.)[:,:,nax]
    def nabla3_x_log_pdf(self, x, *args, **kwargs):
        m = self.mu
        b = self.beta
        z = (x-m)/b
        return (np.exp(-z)/b**3.)[:,:,nax,nax]

class WeibullDistribution(FrozenDistribution_1d):
    def __init__(self, c, mu=0., sigma=1.):
        super(WeibullDistribution,self).__init__()
        self.c = c
        self.mu = mu
        self.sigma=sigma
        self.dist = stats.weibull_min(c=self.c, loc=self.mu, scale=self.sigma)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf(x).flatten()
    def grad_x_log_pdf(self, x, *args, **kwargs):
        c = self.c
        m = self.mu
        s = self.sigma
        out = (c-1.)/(x-m) - c/s * ((x-m)/s)**(c-1.)
        return out
    def hess_x_log_pdf(self, x, *args, **kwargs):
        c = self.c
        m = self.mu
        s = self.sigma
        out = - (c-1.)/(x-m)**2. - (c*(c-1.))/s**2. * ((x-m)/s)**(c-2.)
        return out[:,:,nax]

class BananaDistribution(PushForwardTransportMapDistribution):
    def __init__(self, a, b, mu, sigma2):
        import TransportMaps.Maps as MAPS
        gauss_map = MAPS.LinearTransportMap( mu, npla.cholesky(sigma2) )
        ban_map = MAPS.FrozenBananaTransportMap(a, b)
        tm = MAPS.CompositeTransportMap(ban_map, gauss_map)
        base_distribution = StandardNormalDistribution(2)
        super(BananaDistribution, self).__init__(tm, base_distribution)




#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import warnings
import numpy as np

from TransportMaps import mpi_eval
from TransportMaps.Distributions.ParametricDistributionBase import *

__all__ = ['TransportMapDistribution',
           'PushForwardTransportMapDistribution',
           'PullBackTransportMapDistribution']

class TransportMapDistribution(ParametricDistribution):
    r""" Abstract class for densities of the transport map type (:math:`T^\sharp \pi` or :math:`T_\sharp \pi`)

    Args:
      transport_map (Maps.TriangularTransportMap): transport map :math:`T`
      base_distribution (Distributions.Distribution): distribution :math:`\pi`

    .. seealso:: :class:`PushForwardTransportMapDistribution` and :class:`PullBackTransportMapDistribution`.
    """

    def __init__(self, transport_map, base_distribution):
        if transport_map.dim != base_distribution.dim:
            raise ValueError("The transport_map and the base_distribution should have " +
                             "the same dimension")
        super(TransportMapDistribution,self).__init__(transport_map.dim)
        self.transport_map = transport_map
        self.base_distribution = base_distribution

    @property
    def coeffs(self):
        r""" Get the coefficients :math:`{\bf a}` of the distribution

        .. seealso:: :func:`ParametricDistribution.coeffs`
        """
        return self.transport_map.coeffs

    @property
    def n_coeffs(self):
        r""" Get the number :math:`N` of coefficients
        
        .. seealso:: :func:`ParametricDistribution.n_coeffs`
        """
        return self.transport_map.n_coeffs

    @coeffs.setter
    def coeffs(self, coeffs):
        r""" Set the coefficients :math:`{\bf a}` of the distribution

        .. seealso:: :func:`ParametricDistribution.coeffs`
        """
        self.transport_map.coeffs = coeffs

    def rvs(self, m, mpi_pool=None, batch_size=None):
        r""" Generate :math:`m` samples from the distribution.

        Args:
          m (int): number of samples to generate
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes
          batch_size (int): whether to generate samples in batches

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- :math:`m`
             :math:`d`-dimensional samples
        """
        if batch_size is None:
            batch_size = m
        samples = np.zeros((m, self.dim))
        niter = m // batch_size + (1 if m % batch_size > 0 else 0)
        for it in range(niter):
            nnew = min(m-it*batch_size, batch_size)
            samp_start = it * batch_size
            samp_stop = samp_start + nnew
            # Sample
            base_samples = self.base_distribution.rvs(nnew, mpi_pool=mpi_pool)
            samples[samp_start:samp_stop,:] = self.map_samples_base_to_target(
                base_samples, mpi_pool=mpi_pool)
        return samples

    def quadrature(self, qtype, qparams, mpi_pool=None):
        r""" Generate quadrature points and weights.

        Args:
          qtype (int): quadrature type number. The different types are defined in
            the associated sub-classes.
          qparams (object): inputs necessary to the generation of the selected
            quadrature
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes

        Return:
          (:class:`tuple` (:class:`ndarray<numpy.ndarray>` [:math:`m,d`],
            :class:`ndarray<numpy.ndarray>` [:math:`m`])) -- list of quadrature
            points and weights
        """
        (base_x,w) = self.base_distribution.quadrature(qtype, qparams, mpi_pool=mpi_pool)
        x = self.map_samples_base_to_target(base_x, mpi_pool=mpi_pool)
        return (x,w)

    def map_samples_base_to_target(self, x, mpi_pool=None):
        r""" [Abstract] Map input samples (assumed to be from :math:`\pi`) to the corresponding samples from :math:`T^\sharp \pi` or :math:`T_\sharp \pi`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): input samples
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- corresponding samples
        """
        raise NotImplementedError("Abstract method. Implement in sub-class.")

    def map_samples_target_to_base(self, x, mpi_pool=None):
        r""" [Abstract] Map input samples (assumed to be from :math:`T^\sharp \pi` or :math:`T_\sharp \pi`) to the corresponding samples from :math:`\pi`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): input samples
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- corresponding samples
        """
        raise NotImplementedError("Abstract method. Implement in sub-class.")


class PushForwardTransportMapDistribution(TransportMapDistribution):
    r""" Class for densities of the transport map type :math:`T_\sharp \pi`

    Args:
      transport_map (Maps.TriangularTransportMap): transport map :math:`T`
      base_distribution (Distributions.Distribution): distribution :math:`\pi`

    .. seealso:: :class:`TransportMapDistribution`
    """

    def pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`T_\sharp \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of :math:`T_\sharp \pi`
            at the ``x`` points.
        """
        return np.exp( self.log_pdf(x, params, idxs_slice) )
        
    def log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\log T_\sharp \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of :math:`\log T_\sharp\pi`
            at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.log_pushforward(self.base_distribution, x,
                                                  params_t, params_pi, idxs_slice)

    def grad_a_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla_{\bf a} \log T_\sharp \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of
            :math:`\nabla_{\bf a} \log T_\sharp \pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.grad_a_log_pushforward(self.base_distribution, x,
                                                         params_t, params_pi, idxs_slice)

    def tuple_grad_a_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\left(\log T_\sharp \pi({\bf x}), \nabla_{\bf a} \log T_\sharp \pi({\bf x})\right)`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`tuple`) --
            :math:`\left(\log T_\sharp \pi({\bf x}), \nabla_{\bf a} \log T_\sharp \pi({\bf x})\right)`
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.tuple_grad_a_log_pushforward(self.base_distribution, x,
                                                               params_t, params_pi, idxs_slice)
        
    def grad_x_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla_{\bf x} \log \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- values of
            :math:`\nabla_x\log\pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.grad_x_log_pushforward(self.base_distribution, x,
                                                         params_t, params_pi, idxs_slice)

    def tuple_grad_x_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\left(\log \pi({\bf x}), \nabla_{\bf x} \log \pi({\bf x})\right)`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`tuple`) --
            :math:`\left(\log \pi({\bf x}), \nabla_{\bf x} \log \pi({\bf x})\right)`
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.tuple_grad_x_log_pushforward(self.base_distribution, x,
                                                               params_t, params_pi, idxs_slice)

    def hess_x_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla^2_{\bf x} \log \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) -- values of
            :math:`\nabla^2_x\log\pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.hess_x_log_pushforward(self.base_distribution, x,
                                                         params_t, params_pi, idxs_slice)
        
    def map_samples_base_to_target(self, x, mpi_pool=None):
        r""" Map input samples (assumed to be from :math:`\pi`) to the corresponding samples from :math:`T_\sharp \pi`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): input samples
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- corresponding samples
        """
        scatter_tuple = (['x'], [x])
        out = mpi_eval("evaluate", scatter_tuple=scatter_tuple, obj=self.transport_map,
                       mpi_pool=mpi_pool)
        return out

    def map_samples_target_to_base(self, x, mpi_pool=None):
        r""" Map input samples assumed to be from :math:`T_\sharp \pi` to the corresponding samples from :math:`\pi`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): input samples
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- corresponding samples
        """
        scatter_tuple = (['x'], [x])
        out = mpi_eval("inverse", scatter_tuple=scatter_tuple, obj=self.transport_map,
                       mpi_pool=mpi_pool)
        return out

    def minimize_kl_divergence(self, tar, qtype, qparams,
                               parbase=None, partar=None,
                               x0=None,
                               regularization=None,
                               tol=1e-4, maxit=100, ders=2,
                               fungrad=False,
                               batch_size=None,
                               mpi_pool=None,
                               grad_check=False, hess_check=False):
        r""" Solve :math:`\arg \min_{\bf a}\mathcal{D}_{KL}\left((T_\sharp \pi)_{\bf a}, \pi_{\rm tar}\right)`

        The minimization is not directly done on the original problem,
        but on the equivalent problem

        .. math::

           \arg \min_{\bf a}\mathcal{D}_{KL}\left(\pi, (T^\sharp\pi_{\rm tar})_{\bf a}\right) \;.
        
        Args:
          tar (:class:`Distribution<TransportMaps.Distribution>`): target distribution
            :math:`\pi_{\rm tar}`
          qtype (int): quadrature type number provided by :math:`\pi`
          qparams (object): inputs necessary to the generation of the selected
            quadrature
          partar (dict): parameters for the evaluation of :math:`\pi_{\rm tar}`
          parbase (dict): parameters for the evaluation of :math:`\pi`
          x0 (:class:`ndarray<numpy.ndarray>` [:math:`N`]): coefficients to be used
            as initial values for the optimization
          regularization (dict): defines the regularization to be used.
            If ``None``, no regularization is applied.
            If key ``type=='L2'`` then applies Tikonhov regularization with
            coefficient in key ``alpha``.
          tol (float): tolerance to be used to solve the KL-divergence problem.
          maxit (int): maximum number of iterations
          ders (int): order of derivatives available for the solution of the
            optimization problem. 0 -> derivative free, 1 -> gradient, 2 -> hessian.
          fungrad (bool): whether the two distributions :math:`\pi_1,\pi_2` provide the method
            :func:`Distribution.tuple_grad_x_log_pdf` computing the evaluation and the
            gradient in one step. This is used only for ``ders==1``.
          batch_size (:class:`list<list>` [3 or 2] of :class:`int<int>`): the list contains the
            size of the batch to be used for each iteration. A size ``1`` correspond
            to a completely non-vectorized evaluation. A size ``None`` correspond to a
            completely vectorized one.
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes
          grad_check (bool): whether to use finite difference to check the correctness of
            of the gradient
          hess_check (bool): whether to use finite difference to check the correctenss of
            the Hessian

        Returns:
          log (dict): log informations from the solver
        """
        pbdistribution = PullBackTransportMapDistribution(self.transport_map, tar)
        par_base = parbase
        par_pbdistribution = partar
        # par_pbdistribution = {'params_t': {},
        #                  'params_pi': partar}
        log = self.transport_map.minimize_kl_divergence(
            self.base_distribution, pbdistribution, qtype=qtype, qparams=qparams,
            params_d1=par_base, params_d2=par_pbdistribution,
            x0=x0, regularization=regularization,
            tol=tol, maxit=maxit, ders=ders, fungrad=fungrad, batch_size=batch_size, 
            mpi_pool=mpi_pool, grad_check=grad_check, hess_check=hess_check)
        return log

class PullBackTransportMapDistribution(TransportMapDistribution):
    r""" Class for densities of the transport map type :math:`T^\sharp \pi`

    Args:
      transport_map (Maps.TriangularTransportMap): transport map :math:`T`
      base_distribution (Distributions.Distribution): distribution :math:`\pi`

    .. seealso:: :class:`TransportMapDistribution`
    """

    def pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`T^\sharp \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of :math:`T^\sharp \pi`
            at the ``x`` points.
        """
        return np.exp( self.log_pdf(x, params, idxs_slice))

    def log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\log T^\sharp \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of :math:`\log T^\sharp \pi`
            at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.log_pullback(self.base_distribution, x,
                                               params_t, params_pi, idxs_slice)

    def grad_x_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla_{\bf x} \log T^\sharp \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of
            :math:`\nabla_{\bf x} \log T^\sharp \pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.grad_x_log_pullback(self.base_distribution, x,
                                                      params_t, params_pi, idxs_slice)

    def tuple_grad_x_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\left(\log T^\sharp \pi({\bf x}), \nabla_{\bf x} \log T^\sharp \pi({\bf x})\right)`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`tuple`) --
            :math:`\left(\log T^\sharp \pi({\bf x}), \nabla_{\bf x} \log T^\sharp \pi({\bf x})\right)`
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.tuple_grad_x_log_pullback(self.base_distribution, x,
                                                            params_t, params_pi, idxs_slice)

    def hess_x_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla^2_{\bf x} \log T^\sharp \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of
            :math:`\nabla^2_{\bf x} \log T^\sharp \pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.hess_x_log_pullback(self.base_distribution, x,
                                                      params_t, params_pi, idxs_slice)

    def grad_a_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla_{\bf a} \log T^\sharp \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          
        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of
            :math:`\nabla_{\bf a} \log T^\sharp \pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.grad_a_log_pullback(self.base_distribution, x,
                                                      params_t, params_pi, idxs_slice)

    def tuple_grad_a_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\left(\log T^\sharp \pi({\bf x}), \nabla_{\bf a} \log T^\sharp \pi({\bf x})\right)`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          
        Returns:
          (:class:`tuple`) --
            :math:`\left(\log T^\sharp \pi({\bf x}), \nabla_{\bf a} \log T^\sharp \pi({\bf x})\right)`
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.tuple_grad_a_log_pullback(self.base_distribution, x,
                                                            params_t, params_pi, idxs_slice)

    def hess_a_log_pdf(self, x, params=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla^2_{\bf a} \log T^\sharp \pi({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          
        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of
            :math:`\nabla^2_{\bf a} \log T^\sharp \pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_t = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_t = None
        return self.transport_map.hess_a_log_pullback(self.base_distribution, x,
                                                      params_t, params_pi, idxs_slice)

    def map_samples_base_to_target(self, x, mpi_pool=None):
        r""" Map input samples (assumed to be from :math:`\pi`) to the corresponding samples from :math:`T^\sharp \pi`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): input samples
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- corresponding samples
        """
        scatter_tuple = (['x'], [x])
        out = mpi_eval("inverse", scatter_tuple=scatter_tuple, obj=self.transport_map,
                       mpi_pool=mpi_pool)
        return out

    def map_samples_target_to_base(self, x, mpi_pool=None):
        r""" Map input samples assumed to be from :math:`T^\sharp \pi` to the corresponding samples from :math:`\pi`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): input samples
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- corresponding samples
        """
        scatter_tuple = (['x'], [x])
        out = mpi_eval("evaluate", scatter_tuple=scatter_tuple, obj=self.transport_map,
                       mpi_pool=mpi_pool)
        return out

    def minimize_kl_divergence(self, tar, qtype, qparams,
                               parbase=None, partar=None,
                               x0=None,
                               regularization=None,
                               tol=1e-4, maxit=100, ders=2,
                               fungrad=False,
                               batch_size=None,
                               mpi_pool=None, 
                               grad_check=False, hess_check=False):
        r""" Solve :math:`\arg \min_{\bf a}\mathcal{D}_{KL}\left((T^\sharp \pi)_{\bf a}, \pi_{\rm tar}\right)`

        The minimization is not directly done on the original problem,
        but on the equivalent problem

        .. math::

           \arg \min_{\bf a}\mathcal{D}_{KL}\left(\pi, (T_\sharp\pi_{\rm tar})_{\bf a}\right) \;.
        
        Args:
          tar (:class:`Distribution<TransportMaps.Distribution>`): target distribution
            :math:`\pi_{\rm tar}`
          qtype (int): quadrature type number provided by :math:`\pi`
          qparams (object): inputs necessary to the generation of the selected
            quadrature
          partar (dict): parameters for the evaluation of :math:`\pi_{\rm tar}`
          parbase (dict): parameters for the evaluation of :math:`\pi`
          x0 (:class:`ndarray<numpy.ndarray>` [:math:`N`]): coefficients to be used
            as initial values for the optimization
          regularization (dict): defines the regularization to be used.
            If ``None``, no regularization is applied.
            If key ``type=='L2'`` then applies Tikonhov regularization with
            coefficient in key ``alpha``.
          tol (float): tolerance to be used to solve the KL-divergence problem.
          maxit (int): maximum number of iterations
          ders (int): order of derivatives available for the solution of the
            optimization problem. 0 -> derivative free, 1 -> gradient, 2 -> Hessian
          fungrad (bool): whether the distributions :math:`\pi_1,\pi_2` provide the method
            :func:`Distribution.tuple_grad_x_log_pdf` computing the evaluation and the
            gradient in one step. This is used only for ``ders==1``.
          batch_size (:class:`list<list>` [3 or 2] of :class:`int<int>`): the list
            contains the
            size of the batch to be used for each iteration. A size ``1`` correspond
            to a completely non-vectorized evaluation. A size ``None`` correspond to a
            completely vectorized one.
          mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processes
          grad_check (bool): whether to use finite difference to check the correctness of
            of the gradient
          hess_check (bool): whether to use finite difference to check the correctenss of
            the Hessian

        Returns:
          log (dict): log informations from the solver
        """
        pfdistribution = PushForwardTransportMapDistribution(self.transport_map, tar)
        par_base = parbase
        par_pfdistribution = partar
        # par_pfdistribution = {'params_t': {},
        #                  'params_pi': partar}
        log = self.transport_map.minimize_kl_divergence(
            self.base_distribution, pfdistribution,
            qtype=qtype, qparams=qparams,
            params_d1=par_base, params_d2=par_pfdistribution,
            x0=x0, regularization=regularization,
            tol=tol, maxit=maxit, ders=ders,
            fungrad=fungrad,
            batch_size=batch_size, cache_level=0,
            mpi_pool=mpi_pool, 
            grad_check=grad_check, hess_check=hess_check)
        return log


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import warnings
import numpy as np
import scipy.linalg as scila

from TransportMaps.Functionals.FrozenMonotonicFunctions import \
    FrozenLinear, FrozenExponential, FrozenGaussianToUniform

from TransportMaps.Maps.TriangularTransportMapBase import *

__all__ = ['FrozenLinearDiagonalTransportMap',
           'FrozenExponentialDiagonalTransportMap',
           'FrozenGaussianToUniformDiagonalTransportMap',
           'FrozenBananaTransportMap']

nax = np.newaxis

class FrozenLinearDiagonalTransportMap(MonotonicTriangularTransportMap):
    r""" Linear diagonal transport map :math:`(x_1,\ldots,x_d) \rightarrow (a_1+b_1 x_1, \ldots, a_d + b_d x_d)`

    Args:
       a (:class:`ndarray<numpy.ndarray>` [:math:`d`]): coefficients
       b (:class:`ndarray<numpy.ndarray>` [:math:`d`]): coefficients

    .. note:: This map is frozen, meaning that optimizing the coefficients with
       respect to a certain cost function is not allowed.

    .. seealso:: :class:`TriangularTransportMap` for a description of the
       overridden methods and :class:`FunctionalApproximations.FrozenLinear`
       for a description of the linear approximation used for each component.
    """

    def __init__(self, a, b):
        if len(a) != len(b):
            raise ValueError("Dimension mismatch")
        dim = len(a)
        self.a = a
        self.b = b
        approx_list = []
        active_vars = []
        for d in range(dim):
            approx_list.append( FrozenLinear(1, a[d], b[d]) )
            active_vars.append( [d] )
        super(FrozenLinearDiagonalTransportMap, self).__init__(active_vars, approx_list)

    @property
    def n_coeffs(self):
        return len(self.a)*2

    def set_coeffs(self, a, b):
        raise NotImplementedError("This is a frozen transport map")

    def evaluate(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        xout = self.a[nax,:] + self.b[nax,:] * x
        return xout

    def grad_x(self, x, *args, **kwargs):
        r""" This is a diagonal map """
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        out = np.zeros( (x.shape[0], self.dim, self.dim) )
        out += np.diag(self.b)[nax,:,:]
        return out

    def det_grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return np.prod( self.b )

    def log_det_grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return np.sum( np.log(self.b) )

    def grad_x_log_det_grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return np.zeros((x.shape[0], self.dim))

    def hess_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return np.zeros( (x.shape[0], self.dim, self.dim, self.dim) )

    def hess_x_log_det_grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return np.zeros((x.shape[0], self.dim, self.dim))

    def regression(self, x, w, F, REG, tol):
        r"""
        Raises:
           NotImplementedError: this is a frozen transport map.
        """
        raise NotImplementedError("This is a frozen transport map.")

class FrozenExponentialDiagonalTransportMap(MonotonicTriangularTransportMap):
    r""" Exponential diagonal transport map :math:`(x_1,\ldots,x_d) \rightarrow (\exp(x_1), \ldots, \exp(x_d))`

    Args:
       dim (int): dimension :math:`d` of the transport map

    .. note:: This map is frozen, meaning that optimizing the coefficients with
       respect to a certain cost function is not allowed.

    .. seealso:: :class:`TriangularTransportMap` for a description of the
       overridden methods and
       :class:`FunctionalApproximations.FrozenExponential`
       for a description of the exponential approximation used for each component.
    """
    
    def __init__(self, dim):
        approx_list = []
        active_vars = []
        for d in range(dim):
            approx_list.append( FrozenExponential(1) )
            active_vars.append( [d] )
        super(FrozenExponentialDiagonalTransportMap,self).__init__(active_vars,
                                                                   approx_list)
    @property
    def n_coeffs(self):
        return len(self.a)*3

    def set_coeffs(self, dim):
        pass

    def evaluate(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        xout = np.exp(x)
        return xout

    def grad_x(self, x, *args, **kwargs):
        r""" This is a diagonal map """
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        out = np.zeros( (x.shape[0], self.dim, self.dim) )
        diag_out = np.einsum('...ii->...i', out) # Read/write from Numpy 1.10
        diag_out[:] = self.evaluate(x)
        return out

    def det_grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return np.exp( self.log_det_grad_x(x, *args, **kwargs) )

    def log_det_grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return np.sum( x, axis=1 )

    def grad_x_log_det_grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return np.ones((x.shape[0], self.dim))

    def hess_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        out = np.zeros( (x.shape[0], self.dim, self.dim, self.dim) )
        diag_out = np.einsum('...iii->...i', out) # Read/write from Numpy 1.10
        diag_out[:] = self.evaluate(x)
        return out

    def hess_x_log_det_grad_x(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return np.zeros((x.shape[0], self.dim, self.dim))

    def inverse(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        xout = np.log(x)
        return xout

    def grad_x_inverse(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        xout = 1/x
        return xout

    def hess_x_inverse(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        out = np.zeros( (x.shape[0], self.dim, self.dim, self.dim) )
        diag_out = np.einsum('...iii->...i', out) # Read/write from Numpy 1.10
        diag_out[:] = -1/x**2
        return xout

    def log_det_grad_x_inverse(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        xout = - np.sum( np.log(x), axis=1 )
        return xout

    def grad_x_log_det_grad_x_inverse(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        xout = - 1./x
        return xout

    def hess_x_log_det_grad_x_inverse(self, x, *args, **kwargs):
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        out = np.zeros( (x.shape[0], self.dim, self.dim, self.dim) )
        diag_out = np.einsum('...iii->...i', out) # Read/write from Numpy 1.10
        diag_out[:] = 1/x**2
        return xout

    def regression(self, x, w, F, REG, tol):
        r"""
        Raises:
           NotImplementedError: this is a frozen transport map.
        """
        raise NotImplementedError("This is a frozen transport map.")

class FrozenGaussianToUniformDiagonalTransportMap(MonotonicTriangularTransportMap):
    r""" Gaussian to Uniform diagonal transport map :math:`(x_1,\ldots,x_d) \rightarrow (\frac{1}{2}[1+{\rm erf}(x_1/\sqrt{2})], \ldots, \frac{1}{2}[1+{\rm erf}(x_d/\sqrt{2})])`

    Args:
       dim (int): dimension :math:`d` of the transport map

    .. note:: This map is frozen, meaning that optimizing the coefficients with
       respect to a certain cost function is not allowed.

    .. seealso:: :class:`TriangularTransportMap` for a description of the
       overridden methods and
       :class:`FunctionalApproximations.FrozenGaussianToUniform`
       for a description of the Gaussian to Uniform approximation used for
       each component.
    """
    
    def __init__(self, dim):
        approx_list = []
        active_vars = []
        for d in range(dim):
            approx_list.append( FrozenGaussianToUniform(d+1) )
            active_vars.append( [d] )
        super(FrozenGaussianToUniformDiagonalTransportMap,self).__init__(active_vars,
                                                                         approx_list)

    @property
    def n_coeffs(self):
        return len(self.a)*3

    def set_coeffs(self, dim):
        pass

    def regression(self, x, w, F, REG, tol):
        r"""
        Raises:
           NotImplementedError: this is a frozen transport map.
        """
        raise NotImplementedError("This is a frozen transport map.")

# class FrozenBananaTransportMap(TriangularTransportMap):
#     def __init__(self, a, b, mu, sigma2):
#         self.dim = 2
#         self.a = a
#         self.b = b
#         self.mu = mu
#         self.sigma2 = sigma2
#         self.chol = scila.cholesky( sigma2, lower = True )
#         self.inv_chol = scila.solve_triangular( self.chol,
#                                             np.eye(self.dim),
#                                             lower=True )
#         self.inv_sigma2 = scila.cho_solve( (self.chol,True),
#                                            np.eye(self.dim) )
#     def Tb(self, x, *args, **kwargs):
#         return self.mu + np.dot( self.chol, x.T ).T
#     def Tb_inv(self, x, *args, **kwargs):
#         return scila.solve_triangular( self.chol, (x-self.mu).T, lower=True ).T
#     def Tt(self, x, *args, **kwargs):
#         a = self.a
#         b = self.b
#         y = np.zeros(x.shape)
#         y[:,0] = a * x[:,0]
#         # y[:,1] = x[:,1] / a - b * ((a*x[:,0])**2. + a**2.)
#         y[:,1] = x[:,1] / a - b * ((a*x[:,0])**2.)
#         return y
#     def Tt_inv(self, y, *args, **kwargs):
#         a = self.a
#         b = self.b
#         x = np.zeros(y.shape)
#         x[:,0] = y[:,0] / a
#         # x[:,1] = a * (y[:,1] + b * (y[:,0]**2. + a**2.))
#         x[:,1] = a * (y[:,1] + b * (y[:,0]**2.))
#         return x
#     def evaluate(self, x, *args, **kwargs):
#         return self.Tt( self.Tb(x, par), par )
#     def __call__(self, x):
#         return self.evaluate(x)
#     def inverse(self, x, *args, **kwargs):
#         return self.Tb_inv( self.Tt_inv(x, par), par )
#     def log_det_grad_x(self, x, *args, **kwargs):
#         return 2. * np.log(self.a)

class FrozenBananaTransportMap(TriangularTransportMap):
    def __init__(self, a, b):
        self.dim = self.dim_in = self.dim_out = 2
        self.a = a
        self.b = b
    def evaluate(self, x, *args, **kwargs):
        a = self.a
        b = self.b
        y = np.zeros(x.shape)
        y[:,0] = a * x[:,0]
        y[:,1] = x[:,1] / a - b * ((a*x[:,0])**2. + a**2.)
        # y[:,1] = x[:,1] / a - b * ((a*x[:,0])**2.)
        return y
    def inverse(self, y, *args, **kwargs):
        a = self.a
        b = self.b
        x = np.zeros(y.shape)
        x[:,0] = y[:,0] / a
        x[:,1] = a * (y[:,1] + b * (y[:,0]**2. + a**2.))
        # x[:,1] = a * (y[:,1] + b * (y[:,0]**2.))
        return x
    def log_det_grad_x(self, x, *args, **kwargs):
        return np.ones(x.shape[0])
    def log_det_grad_x_inverse(self, x, *args, **kwargs):
        return np.ones(x.shape[0])
    def grad_x_log_det_grad_x_inverse(self, x, *args, **kwargs):
        return np.zeros(x.shape)
    def hess_x_log_det_grad_x_inverse(self, x, *args, **kwargs):
        return np.zeros((x.shape[0],self.dim,self.dim))
    def grad_x(self, x, *args, **kwargs):
        out = np.zeros( (x.shape[0], self.dim, self.dim) )
        out[:,0,0] = self.a
        out[:,1,0] = - 2. * self.a**2. * self.b * x[:,0]
        out[:,1,1] = 1./self.a
        return out
    def hess_x(self, x, *args, **kwargs):
        out = np.zeros( (x.shape[0], self.dim, self.dim, self.dim) )
        out[:,1,0,0] = - 2. * self.a**2. * self.b
        return out
    def grad_x_inverse(self, x, *args, **kwargs):
        out = np.zeros( (x.shape[0], self.dim, self.dim) )
        out[:,0,0] = 1./self.a
        out[:,1,0] = 2. * self.a * self.b * x[:,0]
        out[:,1,1] = self.a
        return out
    def hess_x_inverse(self, x, *args, **kwargs):
        out = np.zeros( (x.shape[0], self.dim, self.dim, self.dim) )
        out[:,1,0,0] = 2. * self.a * self.b
        return out


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

from __future__ import print_function

import sys
import getopt
import os
import shutil
import time
import datetime
import dill

from TransportMaps.ObjectBase import TMO

__all__ = ['Script']

class Script(TMO):

    def usage(self):
        raise NotImplementedError("To be implemented in subclasses")

    def description(self):
        raise NotImplementedError("To be implemented in subclasses")

    def full_usage(self):
        self.usage()

    def full_doc(self):
        self.full_usage()
        self.description()

    def tstamp_print(self, msg, *args, **kwargs):
        tstamp = datetime.datetime.fromtimestamp(
            time.time()
        ).strftime('%Y-%m-%d %H:%M:%S')
        print(tstamp + " " + msg, *args, **kwargs)

    def filter_tstamp_print(self, msg, *args, **kwargs):
        if self.VERBOSE:
            self.tstamp_print(msg, *args, **kwargs)

    def filter_print(self, *args, **kwargs):
        if self.VERBOSE:
            print(*args, **kwargs)

    def safe_store(self, data, fname):
        # Backup copy
        shutil.copyfile(fname, fname + '.bak')
        # Store data
        with open(fname, 'wb') as out_stream:
            dill.dump(data, out_stream);
        # Remove backup
        os.remove(fname + '.bak')

    @property
    def short_options(self):
        return "hvI"

    @property
    def long_options(self):
        return [
            # I/O
            "input=", "output=",
            # Logging
            "log=",
            # Computation
            "nprocs="
        ]

    def load_opts(self, opts):
        for opt, arg in opts:
            if opt == '-h':
                self.full_doc()
                sys.exit()

            # Verbose
            elif opt == '-v':
                self.VERBOSE = True

            # Interactive
            elif opt in ("-I"):
                self.INTERACTIVE = True

            # I/O
            elif opt in ("--input"):
                self.INPUT = arg
            elif opt in ("--output"):
                self.OUTPUT = arg

            # Logging
            elif opt in ['--log']:
                self.LOGGING_LEVEL = int(arg)

            # Computation
            elif opt in ("--nprocs"):
                self.NPROCS = int(arg)
        
    def __init__(self, argv):
        super(Script, self).__init__()

        # Options
        self.INTERACTIVE = False
        # I/O
        self.INPUT = None
        self.OUTPUT = None
        # Logging
        self.VERBOSE = False
        self.LOGGING_LEVEL = 30 # Warnings
        # Parallelization
        self.NPROCS = 1

        # Parse
        try:
            opts, args = getopt.getopt(
                argv, self.short_options, self.long_options)
        except getopt.GetoptError as e:
            self.full_usage()
            raise e

        self.load_opts(opts)

        # Check for required arguments
        if None in [self.INPUT, self.OUTPUT]:
            self.usage()
            self.tstamp_print("ERROR: Option --input and --output must be specified")
            sys.exit(3)


    def load(self):
        raise NotImplementedError("To be implemented in sub-classes")

    def run(self, mpi_pool):
        raise NotImplementedError("To be implemented in sub-classes")


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

__all__ = []

from ._version import __version__

# Transport map imports
from . import External
from .External import *

from . import ObjectBase
from .ObjectBase import *

from . import Misc
from .Misc import *

from . import XML

from . import FiniteDifference

from . import Distributions
from . import Functionals
from . import Maps
from . import Likelihoods

from . import Routines
from .Routines import *

from . import Defaults
from .Defaults import *

from . import Algorithms
# from . import Optimization
from . import Diagnostics
from . import Samplers
from . import CLI
from . import tests

__all__ += External.__all__
__all__ += ObjectBase.__all__
__all__ += Misc.__all__
__all__ += ['FiniteDifference']
__all__ += ['Distributions']
__all__ += ['Functionals']
__all__ += ['Likelihoods']
__all__ += ['Maps']
__all__ += Routines.__all__
__all__ += Defaults.__all__
__all__ += ['Algorithms']
# __all__ += ['Optimization']
__all__ += ['Diagnostics']
__all__ += ['Samplers']
__all__ += ['XML']
__all__ += ['CLI']
__all__ += ['tests']

############
# DEPRECATED
from . import Densities
__all__ += ['Densities']
############

XML_NAMESPACE = '{TransportMaps}'

__author__ = "Transport Map Team"
__copyright__ = """LGPLv3, Copyright (C) 2015-2017, Massachusetts Institute of Technology"""
__credits__ = ["Transport Map Team"]
__maintainer__ = "Transport Map Team"
__website__ = "transportmaps.mit.edu"
__status__ = "Development"


##############################
# Linking to SpectralToolbox
def linking():
    import os.path
    tm_dir = os.path.dirname(os.path.realpath(__file__))
    if not os.path.islink(tm_dir + '/XML/schema/basis.xsd'):
        import warnings
        import os
        import SpectralToolbox.Spectral1D as S1D
        warnings.warn("Re-linking to SpectralToolbox. " + \
                      "This is done every time a new version of either " + \
                      "SpectralToolbox or TransportMaps are updated.",
                      UserWarning)
        s1d_dir = os.path.dirname(os.path.realpath(S1D.__file__))
        os.symlink(s1d_dir + '/XML/schema/basis.xsd',
                   tm_dir + '/XML/schema/basis.xsd')
linking()


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np
import numpy.linalg as npla

from .AdaptivityBase import Builder
from TransportMaps.Distributions import \
    PushForwardTransportMapDistribution, PullBackTransportMapDistribution
from TransportMaps.Routines import laplace_approximation
from TransportMaps.Diagnostics.Routines import variance_approx_kl
from TransportMaps.Maps.FullTransportMaps import LinearTransportMap

__all__ = ['KullbackLeiblerBuilder',
           'SequentialKullbackLeiblerBuilder',
           'ToleranceSequentialKullbackLeiblerBuilder']

class KullbackLeiblerBuilder(Builder):
    r""" Basis builder through minimization of kl divergence

    Given distribution :math:`\nu_\rho` and :math:`\nu_\pi`,
    and the parametric transport map :math:`T[{\bf a}]`,
    provides the functionalities to solve the problem

    .. math::

       \arg\min_{\bf a}\mathcal{D}_{\rm KL}\left(
       T[{\bf a}]_\sharp\rho \Vert \pi\right)

    up to a chosen tolerance.

    Args:
      base_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\rho`
      target_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\pi`
    """
    def __init__(self, base_distribution=None, target_distribution=None):
        self.base_distribution = base_distribution
        self.target_distribution = target_distribution
        super(KullbackLeiblerBuilder, self).__init__()

    @property
    def base_distribution(self):
        return self._base_distribution

    @base_distribution.setter
    def base_distribution(self, dstr):
        self._base_distribution = dstr

    @property
    def target_distribution(self):
        return self._target_distribution

    @target_distribution.setter
    def target_distribution(self, dstr):
        self._target_distribution = dstr

    def solve(self, transport_map, qtype, qparams,
              x0=None,
              regularization=None,
              tol=1e-4, maxit=100, ders=2,
              fungrad=False,
              batch_size=None,
              mpi_pool=None):
        if x0 is None:
            x0 = transport_map.get_default_init_values_minimize_kl_divergence()
        push_base = PushForwardTransportMapDistribution(
            transport_map, self.base_distribution)
        log = push_base.minimize_kl_divergence(
            self.target_distribution,
            qtype=qtype, qparams=qparams,
            x0=x0, regularization=regularization,
            tol=tol, maxit=maxit, ders=ders,
            fungrad=fungrad, batch_size=batch_size,
            mpi_pool=mpi_pool)
        if not log['success']:
            transport_map.coeffs = x0
        return transport_map, log

class SequentialKullbackLeiblerBuilder(KullbackLeiblerBuilder):
    r""" Solve over a list of maps, using the former to warm start the next one

    Given distribution :math:`\nu_\rho` and :math:`\nu_\pi`,
    and the list of parametric transport maps
    :math:`[T_1[{\bf a}_1,\ldots,T_n[{\bf a}_n]`,
    provides the functionalities to solve the problems

    .. math::

       \arg\min_{{\bf a}_i}\mathcal{D}_{\rm KL}\left(
       T_i[{\bf a}_i]_\sharp\rho \Vert \pi\right)

    up to a chosen tolerance, where the numerical solution for map
    :math:`T_{i+1}` is started at :math:`T_i`

    Args:
      base_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\rho`
      target_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\pi`
    """        
    def solve(self, transport_map_list, solve_params_list, regression_params_list=None):
        r"""
        Args:
          transport_map_list (:class:`list` of :class:`TransportMap<TransportMaps.Maps.TransportMap>`):
            list of transport maps :math:`[T_1[{\bf a}_1,\ldots,T_n[{\bf a}_n]`
          solve_params_list (:class:`list` of :class:`dict`): parameters to be used for the solution
          regression_params_list (:class:`list` of :class:`dict`): parameters to be used in the regression step
        Returns:
          (:class:`TransportMaps.Maps.TransportMap`) -- the last transport map fitted.
        """
        if regression_params_list is None:
            regression_params_list = [ {key: sp[key] for key in (
                'qtype', 'qparams', 'regularization', 'tol', 'maxit') if key in sp}
                                       for sp in solve_params_list[1:] ]
        tm, log = super(SequentialKullbackLeiblerBuilder, self).solve(
            transport_map_list[0],
            **solve_params_list[0])
        if not log['success']:
            tm.coeffs = x0
            return tm, log
        for tm_new, solve_params, regression_params in zip(
                transport_map_list[1:], solve_params_list[1:], regression_params_list):
            # fit next map to old map
            tm_old = tm
            regression_params['d'] = self.base_distribution
            tm_new.regression(tm_old, **regression_params)
            # solve for the new map using regressed starting point
            x0 = tm_new.coeffs
            tm, log = super(SequentialKullbackLeiblerBuilder, self).solve(
                tm_new, x0=x0, **solve_params)
            if not log['success']:
                return tm_old, log
        return tm, log

class ToleranceSequentialKullbackLeiblerBuilder(KullbackLeiblerBuilder):
    r""" Solve over a list of maps, using the former to warm start the next one, until a target tolerance is met

    Given distribution :math:`\nu_\rho` and :math:`\nu_\pi`,
    and the list of parametric transport maps
    :math:`[T_1[{\bf a}_1,\ldots,T_n[{\bf a}_n]`,
    provides the functionalities to solve the problems

    .. math::

       \arg\min_{{\bf a}_i}\mathcal{D}_{\rm KL}\left(
       T_i[{\bf a}_i]_\sharp\rho \Vert \pi\right)

    up to a chosen tolerance, where the numerical solution for map
    :math:`T_{i+1}` is started at :math:`T_i`

    Args:
      base_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\rho`
      target_distribution (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        distribution :math:`\nu_\pi`
    """        
    def solve(self, transport_map_list, solve_params_list, regression_params_list=None,
              tol=1e-2, var_diag_params=None):
        r"""
        Args:
          transport_map_list (:class:`list` of :class:`TransportMap<TransportMaps.Maps.TransportMap>`):
            list of transport maps :math:`[T_1[{\bf a}_1,\ldots,T_n[{\bf a}_n]`
          solve_params_list (:class:`list` of :class:`dict`): parameters to be used for the solution
          regression_params_list (:class:`list` of :class:`dict`): parameters to be used in the regression step
          tol (float): tolerance to be met by the variance diagnostic
          var_diag_params (dict): parameters to be used in the variance diagnostic approximation
        Returns:
          (:class:`TransportMaps.Maps.TransportMap`) -- the last transport map fitted.
        """
        if regression_params_list is None:
            regression_params_list = [ {key: sp[key] for key in (
                'qtype', 'qparams', 'regularization', 'tol', 'maxit') if key in sp}
                                       for sp in solve_params_list[1:] ]
        if var_diag_params is None:
            var_diag_params = {'qtype': 0, 'qparams': 1000}
        
        # First find Laplace point and center to it
        lap = laplace_approximation(self.target_distribution)
        lap_map = LinearTransportMap.build_from_Gaussian(lap)

        # Set initial conditions to Laplace approximation
        transport_map_list[0].regression(
            lap_map, d=self.base_distribution,
            qtype=3, qparams=[3]*self.base_distribution.dim,
            regularization={'alpha': 1e-4, 'type': 'L2'})

        x0 = transport_map_list[0].coeffs
        
        # for d, t in enumerate(transport_map_list[0].approx_list):
        #     # Constant part
        #     for i, midx in enumerate(t.c.get_multi_idxs()):
        #         if sum(midx) == 0:
        #             t.c.coeffs[i] = consTerm[d]
        #         elif sum(midx) == 1:
        #             t.c.coeffs[i] = linTerm[d, midx.index(1)]
        #     # Integrated part
        #     for i, midx in enumerate(t.h.get_multi_idxs()):
        #         if sum(midx) == 0:
        #             t.h.coeffs[i] = np.sqrt(linTerm[d,d])
        
        tm, log = super(ToleranceSequentialKullbackLeiblerBuilder, self).solve(
            transport_map_list[0],
            x0 = x0,
            **solve_params_list[0])
        if not log['success']:
            tm.coeffs = x0
            return tm, log
        pull_tar = PullBackTransportMapDistribution(tm, self.target_distribution)
        var = variance_approx_kl(self.base_distribution, pull_tar, **var_diag_params)
        if var <= tol:
            return tm, log
        for tm_new, solve_params, regression_params in zip(
                transport_map_list[1:], solve_params_list[1:], regression_params_list):
            # # fit next map to old map
            # tm_old = tm
            # regression_params['d'] = self.base_distribution
            # tm_new.regression(tm_old, **regression_params)

            tm_old = tm
            for c1, c2 in zip(tm_old.approx_list, tm_new.approx_list):
                # Constant part
                for i1, midx1 in enumerate(c1.c.get_multi_idxs()):
                    for i2, midx2 in enumerate(c2.c.get_multi_idxs()):
                        if midx1 == midx2:
                            break
                    c2.c.coeffs[i2] = c1.c.coeffs[i1]
                # Integrated part
                for i1, midx1 in enumerate(c1.h.get_multi_idxs()):
                    for i2, midx2 in enumerate(c2.h.get_multi_idxs()):
                        if midx1 == midx2:
                            break
                    c2.h.coeffs[i2] = c1.h.coeffs[i1]
            
            # solve for the new map using regressed starting point
            x0 = tm_new.coeffs
            tm, log = super(ToleranceSequentialKullbackLeiblerBuilder, self).solve(
                tm_new, x0=x0, **solve_params)
            if not log['success']:
                return tm_old, log
            pull_tar = PullBackTransportMapDistribution(tm, self.target_distribution)
            var = variance_approx_kl(self.base_distribution, pull_tar, **var_diag_params)
            if var <= tol:
                return tm, log
        # Variance was not met
        log['success'] = False
        log['msg'] = "Desired tolerance was no met by the map adaptivity. " + \
                     "Target variance: %e - Variance: %e " % (tol, var)
        return tm, log


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

__all__ = ['MPI_SUPPORT', 'PLOT_SUPPORT', 'H5PY_SUPPORT', 'PYHMC_SUPPORT']

import warnings

try:
    import mpi_map
    from mpi4py import MPI
except ImportError:
    warnings.warn("MPI support disabled: install mpi4py and mpi_map if needed.")
    MPI_SUPPORT = False
else:
    MPI_SUPPORT = True
    # __all__ += ['mpi_map']

try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
except ImportError:
    PLOT_SUPPORT = False
    warnings.warn("Plotting support disabled: install matplotlib if needed.")
else:
    PLOT_SUPPORT = True
    # __all__ += ['mpl', 'plt']

try:
    import h5py
except ImportError:
    warnings.warn("H5 file support disabled: install h5py if needed.")
    H5PY_SUPPORT = False
else:
    H5PY_SUPPORT = True
    # __all__ += ['h5py']

try:
    import pyhmc
except ImportError:
    warnings.warn("pyhmc is not supported: install pyhmc if needed.")
    PYHMC_SUPPORT = False
else:
    PYHMC_SUPPORT = True